package com.sillycycle.bagleyd.pyraminx;

/*
 * @(#)PyraminxCanvas.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>PyraminxCanvas</code> class does all the graphics for
 * <code>PyraminxFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

/*
The Simple Solutions to Cubic Puzzles by James G. Nourse, Bantam Books,
New York, 1981. pp 8-15.

Mastering the Magic Pyramid by Tom Werneck, Evans Brothers Limited, London,
1981. pp 109-111.

Beyond Rubik's Cube: spheres, pyramids, dodecahedrons and God knows what
else by Douglas R. Hofstadter, Scientific American, July 1982, pp 16-31.

Magic Cubes 1996 Catalog of Dr. Christoph Bandelow.
*/

public class PyraminxCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 100;

	static final int TOP = 0;
	static final int TR = 1;
	static final int RIGHT = 2;
	static final int BOTTOM = 3;
	static final int BL = 4;
	static final int LEFT = 5;
	static final int COORD = 6;
	static final int CW = 7;
	static final int CCW = 11;
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int MAX_VIEWS = 2;
	static final int MAX_FACES = 4;
	static final int MAX_SIDES = (MAX_FACES / MAX_VIEWS);
	static final int MAX_ORIENT = (3 * MAX_SIDES);

	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "pyraminx.dat";
	static final String SYMBOL = ":";

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MAX_FACES];
	int[]       periodOfPosition = null;
	int[]       directionOfPosition = null;
	int[]       startPosition = null;
	int	 currentStyle = -1;
	int	 currentDirection = -1;
	int	 currentFace = -1, nextFace = -1;
	int	 currentPosition = -1, nextPosition = -1;
	int	 currentView = -1, nextView = -1;
	boolean     nextShift = false;
	boolean     currentControl = false, nextControl = false;
	int	 rowPosition;
	boolean     started, cheat, vertical;
	boolean     mono, reverse, orient, sticky, practice;
	int	 delay;
	int	 lastX, lastY, dr;
	int	 size, period, sizeSize;
	Point       coreSize;
	Point       offset, letter;
	Point       puzzleSize, puzzleOffset, offsetD;
	int	 delta, facetLength, faceLength, viewLength;
	int	 sideOffset, orientLineLength, orientDiagLength;
	PyraminxCRD crd = new PyraminxCRD(-1, -1, -1);

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	Random      generator;
	PyraminxStack undo = null, redo = null;
	PyraminxSolve solve = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;

	PyraminxLoc[][] slideNextRowP2 = new PyraminxLoc[MAX_SIDES][3];
	PyraminxNextRowP3[][] slideNextRowP3 =
		new PyraminxNextRowP3[MAX_SIDES][MAX_ORIENT];
	static final int[][] directionNextFace = {
		/* direction, face */
		{3, 0, 1, 2},
		{3, 3, 1, 1},
		{1, 3, 3, 1},
		{1, 2, 3, 0},
		{2, 2, 0, 0},
		{2, 0, 0, 2}
	};
	static final int[][] rotOrientRowP3 = {
		/* current orient, rotation */
		{1, 5, 1, 5, 4, 2},
		{2, 0, 2, 0, 5, 3},
		{3, 1, 3, 1, 0, 4}
	};
	PyraminxLoc[][] facetLoc = null;
	PyraminxLoc[][] startLoc = null;
	PyraminxLoc[][] faceLoc = null;
	PyraminxLoc[][][] rowLoc = null;
	Point[][] triangleUnit = new Point[MAX_SIDES][4];
	int[][] triangleListX = new int[MAX_SIDES][4];
	int[][] triangleListY = new int[MAX_SIDES][4];
	Point[] letterList = new Point[MAX_SIDES];
	Point[] offsetList = new Point[MAX_SIDES];

	@Override
	protected String paramString() {
		return super.paramString() +
		",size=" + size +
		",period=" + period +
		",orient=" + orient +
		",sticky=" + sticky +
		",practice=" + practice +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkPieces() {
		if (size < PyraminxInterface.MIN_FACETS) {
			System.out.println(size +
				" number of pieces on a edge out of bounds, must be at least " +
				PyraminxInterface.MIN_FACETS +
				", defaulting to " +
				PyraminxInterface.DEFAULTFACETS);
			size = PyraminxInterface.DEFAULTFACETS;
		}
		if (period < PyraminxInterface.PERIOD2 ||
			 period > PyraminxInterface.BOTH) {
			System.out.println("Mode " + period +
				" is in error, use 2 for Period2, 3 for Period3, 4 for Both, defaulting to " +
				PyraminxInterface.DEFAULT_MODE);
			period = PyraminxInterface.DEFAULT_MODE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		for (int face = 0; face < MAX_FACES; face++) {
			int tempFace = facetLoc[face][0].face;
			int tempRotation = facetLoc[face][0].rotation;

			for (int position = 1; position < sizeSize;
					position++) {
				if (tempFace != /* MAX_SIDES * view + face */
					 facetLoc[face][position].face ||
					   (orient && tempRotation !=
				      facetLoc[face][position].rotation))
					return false;
			}
		}
		return true;
	}

	/* This is fast for small i, a -1 is returned for negative i. */
	static int SQRT(int i) {
		int j = 0;

		while (j * j <= i)
			j++;
		return (j - 1);
	}

	PyraminxCRD toCRD(int face, int position) {
		int diag, diag2;

		diag = SQRT(position);
		diag2 = diag * diag;
		/* Passing diagonal so there is no sqrt calculation again */
		if ((face & 1) == UP) {
			return (new PyraminxCRD(
				size - 1 - (position - diag2) / 2,
				size - 1 - (diag2 + 2 * diag - position) / 2,
				2 * size - 1 - diag));
		}
		return (new PyraminxCRD(
			(position - diag2) / 2,
			(diag2 + 2 * diag - position) / 2,
			diag));
	}

	int toPosition(PyraminxCRD crdRotate) {
		int diag;

		if (crdRotate.diagonal < size)
			return (crdRotate.diagonal * crdRotate.diagonal
				+ crdRotate.diagonal + crdRotate.column - crdRotate.row);
		diag = 2 * size - 1 - crdRotate.diagonal;
		return (diag * diag + diag + crdRotate.row - crdRotate.column);
	}

	static int positionCRD(int dir, int I, int J, int side) {
		switch (dir) {
		case TOP:
		case BOTTOM:
			return I;
		case RIGHT:
		case LEFT:
			return J;
		default:
			/* dir == TR || dir == BL */
			return (I + J + side);
		}
	}

	int length(int face, int dir, int h) {
		if (dir == TR || dir == BL)
			return ((face == UP) ? 2 * size - 1 - h : h);
		return ((face == UP) ? h : size - 1 - h);
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawOrientLine(Graphics g, int rotation, int dx, int dy,
			int side) {
		int x1 = 0, x2 = 0, y1 = 0, y2 = 0;
		int temp1 = facetLength / 2 + 1;
		int temp2 = facetLength + 1;
		int fix = (size == 1) ? 1 : 0;

		switch (rotation) {
		case TOP:
			x2 = x1 = dx + temp1 + 2;
			y1 = (side == UP) ? dy + temp1 + 1 : dy;
			y2 = y1 + orientLineLength;
			break;
		case TR:
			x1 = (side == UP) ? dx + temp2 + 1 : dx + temp1;
			x2 = x1 - orientDiagLength;
			y1 = (side == UP) ? dy + temp1 + 1 : dy;
			y2 = y1 + orientDiagLength;
			break;
		case RIGHT:
			x1 = (side == UP) ? dx + temp2 + 1 : dx + temp1 - fix -
				1;
			x2 = x1 - orientLineLength;
			y2 = y1 = dy + temp1 - 1;
			break;
		case BOTTOM:
			x2 = x1 = dx + temp1 - 1;
			y1 = (side == UP) ? dy + temp2 + 1 : dy + temp1 - fix -
				1;
			y2 = y1 - orientLineLength;
			break;
		case BL:
			x1 = (side == UP) ? dx + temp1 : dx;
			x2 = x1 + orientDiagLength;
			y1 = (side == UP) ? dy + temp2 + 1 : dy + temp1;
			y2 = y1 - orientDiagLength;
			break;
		case LEFT:
			x1 = (side == UP) ? dx + temp1 + 1 : dx;
			x2 = x1 + orientLineLength;
			y2 = y1 = dy + temp1 + 2;
			break;
		default:
			System.err.println("drawOrientLine: rotation " +
				rotation);
		}
		g.drawLine(x1, y1, x2, y2);
	}

	void drawTriangle(int face, int position, int offsetPosition) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int faceOnView, rotation, side, view, l;
			PyraminxCRD crdRotate = toCRD(face, position);

			view = (face / MAX_SIDES == UP) ? DOWN : UP;
			side = crdRotate.diagonal - crdRotate.column - crdRotate.row;
			Point start = new Point(puzzleOffset.x + delta,
				puzzleOffset.y + delta);
			Color fillColor, lineColor;

			if (!vertical && view == DOWN) {
				faceOnView = (side == UP) ? DOWN : UP;
				start.x += (size - crdRotate.column - 1) *
					(facetLength + delta);
				start.y += (size - crdRotate.row - 1) *
					(facetLength + delta);
				rotation = (facetLoc[face][position].rotation +
					MAX_ORIENT / 2) % MAX_ORIENT;
				if (2 * size - crdRotate.column - crdRotate.row - 2 +
						faceOnView >= size) {
					start.x += sideOffset;
					start.y += sideOffset;
				}
			} else {
				faceOnView = side;
				start.x += crdRotate.column * (facetLength + delta);
				start.y += crdRotate.row * (facetLength + delta);
				rotation = facetLoc[face][position].rotation;
				if (crdRotate.column + crdRotate.row + faceOnView >= size) {
					start.x += sideOffset;
					start.y += sideOffset;
				}
			}
			if (view == DOWN) {
				if (vertical)
					start.y += viewLength - delta - 1;
				else
					start.x += viewLength - delta - 1;
			}
			l = facetLoc[face][position].face;
			if (offsetPosition == 0) {
				if (mono) {
					if (reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = faceColor[l];
					lineColor = borderColor;
				}
			} else {
				if (mono) {
					if (reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = faceColor[l];
					fillColor = borderColor;
				}
			}
			fillPolygon(g, fillColor, lineColor,
				start.x + faceOnView,
				start.y + faceOnView,
				triangleListX[faceOnView],
				triangleListY[faceOnView],
				3);
			start.x += -faceOnView;
			start.y += -faceOnView;
			if (mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = facetLength / 5;
				g.setFont(new Font("arial", Font.PLAIN, fontSize));

				buf = Character.toString(faceNames.charAt(l));
				g.drawString(buf,
					start.x + letterList[faceOnView].x,
					start.y + letterList[faceOnView].y);
			}
			if (orient) {
				g.setColor(lineColor);
				drawOrientLine(g, rotation, start.x, start.y,
					faceOnView);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < sizeSize; position++)
			drawTriangle(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < MAX_FACES; face++)
			drawFace(face);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			Point start = new Point(1 + puzzleOffset.x,
				1 + puzzleOffset.y);
			Point length = new Point(viewLength - delta +
				puzzleOffset.x, viewLength - delta +
				puzzleOffset.y);
			if (show) {
				g.setColor(currentForeground);
				g.drawLine(start.x, start.y, length.x, start.y);
				g.drawLine(start.x, start.y, start.x, length.y);
				g.drawLine(start.x, length.y, length.x, start.y);
				int longlength = 2 * viewLength - 2 * delta - 1;
				if (vertical) {
					longlength += puzzleOffset.y;
					g.drawLine(start.x, length.y,
						start.x, longlength);
					g.drawLine(length.x, length.y,
						length.x, longlength);
					g.drawLine(start.x, longlength,
						length.x, longlength);
					g.drawLine(start.x, longlength,
						length.x, length.y);
					g.drawLine(length.x, start.y,
						length.x, length.y);
					g.setColor(foreground);
					g.drawLine(0, length.y,
						coreSize.x, length.y);
				} else {
					longlength += puzzleOffset.x;
					g.drawLine(length.x, start.y,
						longlength, start.y);
					g.drawLine(length.x, length.y,
						longlength, length.y);
					g.drawLine(longlength, start.y,
						longlength, length.y);
					g.drawLine(longlength, start.y,
						length.x, length.y);
					g.drawLine(start.x, length.y,
						length.x, length.y);
					g.setColor(foreground);
					g.drawLine(length.x, 0,
						length.x, coreSize.y);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void moveNoPieces() {
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_ILLEGAL);
	}

	void rotateFace(int view, int faceOnView, int direction) {
		int face, position;
		PyraminxCRD crdTemp = new PyraminxCRD(0, 0, 0);

		/* Read Face */
		for (int g = 0; g < size; g++)
		  for (int h = 0; h < size - g; h++)
		    for (int side = 0; side < MAX_SIDES; side++)
			if (g + h + side < size) {
				if (faceOnView == DOWN) {
					crdTemp.column = h;
					crdTemp.row = g;
					crdTemp.diagonal = h + g + side;
					face = view * MAX_SIDES +
						((crdTemp.diagonal >= size) ? UP : DOWN);
					position = toPosition(crdTemp);
					faceLoc[side][h + g * size] =
						facetLoc[face][position];
				} else { /* faceOnView == UP */
					crdTemp.column = size - 1 - h;
					crdTemp.row = size - 1 - g;
					crdTemp.diagonal = crdTemp.column + crdTemp.row + ((side == UP) ? DOWN : UP);
					face = view * MAX_SIDES +
						((crdTemp.diagonal >= size) ? UP : DOWN);
					position = toPosition(crdTemp);
					faceLoc[side][h + g * size] =
					facetLoc[face][position];
				}
			}
		/* Write Face */
		if (faceOnView == DOWN) {
		  for (int g = 0; g < size; g++)
		    for (int h = 0; h < size - g; h++)
		      for (int side = 0; side < MAX_SIDES; side++)
			if (g + h + side < size) {
				crdTemp.column = h;
				crdTemp.row = g;
				crdTemp.diagonal = h + g + side;
				face = view * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				if (direction == CCW)
					facetLoc[face][position] =
						faceLoc[side][size - 1 - g - h - side +
						h * size];
				else	/* direction == CW */
					facetLoc[face][position] =
						faceLoc[side][g +
						(size - 1 - g - h - side) * size];
				facetLoc[face][position].rotation =
					(direction == CW) ?
					(facetLoc[face][position].rotation + 2) %
					MAX_ORIENT :
					(facetLoc[face][position].rotation + 4) %
					MAX_ORIENT;
				drawTriangle(face, position, 0);
			}
		} else { /* faceOnView == UP */
		  for (int g = size - 1; g >= 0; g--)
		    for (int h = size - 1; h >= size - 1 - g; h--)
		      for (int side = 1; side >= 0; side--)
			if (g + h + side >= size) {
				crdTemp.column = h;
				crdTemp.row = g;
				crdTemp.diagonal = h + g + side;
				face = view * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				if (direction == CCW)
					facetLoc[face][position] =
						faceLoc[(side == UP) ? DOWN : UP][g + h - size + 1 - ((side == UP) ? DOWN : UP)
						+ (size - 1 - h) * size];
				else /* (direction == CW) */
					facetLoc[face][position] =
						faceLoc[(side == UP) ? DOWN : UP][size - 1 - g +
						(g + h - size + 1 - ((side == UP) ? DOWN : UP)) * size];
				facetLoc[face][position].rotation =
					(direction == CW) ?
					(facetLoc[face][position].rotation + 2) %
					MAX_ORIENT :
					(facetLoc[face][position].rotation + 4) %
					MAX_ORIENT;
				drawTriangle(face, position, 0);
			}
		}
	}

	void readCRD2(int view, int dir, int h, int orientLine) {
		int i, j, faceOnView, s, face, position, v = view;
		PyraminxCRD crdTemp = new PyraminxCRD(0, 0, 0);

	    switch (dir) {
		case TOP:
		case BOTTOM:
		    for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++) {
			    crdTemp.column = h;
			    crdTemp.row = g;
			    crdTemp.diagonal = h + g + side;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    rowLoc[orientLine][side][g] = facetLoc[face][position];
			}
		    break;
		case RIGHT:
		case LEFT:
		    for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++) {
			    crdTemp.column = g;
			    crdTemp.row = h;
			    crdTemp.diagonal = h + g + side;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    rowLoc[orientLine][side][g] =
				    facetLoc[face][position];
			}
		    break;
		default:
		    /* dir == TR || dir == BL */
		    faceOnView = (h < size) ? DOWN : UP;
		    i = (faceOnView == UP) ? size - 1 : 0;
		    j = h % size;
		    for (int g = 0; g < size; g++) {
			for (int side = 0; side < MAX_SIDES; side++) {
			    s = (((side == UP) && (faceOnView == UP)) ||
				    ((side == DOWN) && (faceOnView == DOWN)))
				    ? DOWN : UP;
			    crdTemp.column = i;
			    crdTemp.row = j;
			    crdTemp.diagonal = i + j + s;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    rowLoc[orientLine][side][g] =
				    facetLoc[face][position];
			    if (side == DOWN) {
				if (faceOnView == UP) {
				    if (j == size - 1)
					v = (v == UP) ? DOWN : UP;
				    j = (j + 1) % size;
				} else { /* faceOnView == DOWN */
				    if (j == 0)
					v = (v == UP) ? DOWN : UP;
				    j = (j - 1 + size) % size;
				}
			    }
			}
			i = (faceOnView == UP) ? i - 1 : i + 1;
		    } break;
	    }
	}

	void readCRD3(int view, int faceOnView, int dir, int h, int len,
			int orientLine) {
		int i, j, s, face, position, faceView = faceOnView;
		PyraminxCRD crdTemp = new PyraminxCRD(0, 0, 0);

	    switch (dir) {
		case TOP:
		case BOTTOM:
		    for (int g = 0; g <= len; g++)
			for (int side = 0; side < MAX_SIDES; side++)
			    if ((side == DOWN) || g < len) {
				crdTemp.column = h;
				crdTemp.row = (faceView == UP) ? size - 1 - g : g;
				crdTemp.diagonal = h + crdTemp.row +
					((((side == UP) && (faceView == DOWN)) ||
					((side == DOWN) && (faceView == UP)))
					? UP : DOWN);
				face = view * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				rowLoc[orientLine][side][g] =
					facetLoc[face][position];
			    }
		    break;
		case RIGHT:
		case LEFT:
		    for (int g = 0; g <= len; g++)
			for (int side = 0; side < MAX_SIDES; side++)
			    if ((side == DOWN) || g < len) {
				crdTemp.column = (faceView == UP) ? size - 1 - g : g;
				crdTemp.row = h;
				crdTemp.diagonal = h + crdTemp.column +
					((((side == UP) && (faceView == DOWN)) ||
					((side == DOWN) && (faceView == UP)))
					? UP : DOWN);
				face = view * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				rowLoc[orientLine][side][g] =
					facetLoc[face][position];
			    }
		    break;
		default:
		    /* dir == TR || dir == BL */
		    faceView = (h < size) ? DOWN : UP;
		    i = (faceView == UP) ? size - 1 : 0;
		    j = h % size;
		    for (int g = 0; g <= len; g++) {
			for (int side = 0; side < MAX_SIDES; side++) {
			    if ((side == DOWN) || g < len) {
				s = (((side == UP) && (faceView == UP)) ||
					((side == DOWN) && (faceView == DOWN)))
					? DOWN : UP;
				crdTemp.column = i;
				crdTemp.row = j;
				crdTemp.diagonal = i + j + s;
				face = view * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				rowLoc[orientLine][side][g] =
					facetLoc[face][position];
				if (side == DOWN)
				    j = (faceView == UP) ? j + 1 : j - 1;
			    }
			}
			i = (faceView == UP) ? i - 1 : i + 1;
		    } break;
	    }
	}

	void rotateCRD2(int rotate, int orientLine) {
		for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++)
				rowLoc[orientLine][side][g].rotation =
					(rowLoc[orientLine][side][g].rotation + rotate) % MAX_ORIENT;
	}

	void rotateCRD3(int faceOnView, int dir, int len, int orientLine) {
		int direction, facetOrient;

		for (int g = 0; g <= len; g++)
		  for (int side = 0; side < MAX_SIDES; side++)
			if ((side == DOWN) || g < len) {
				direction = (faceOnView == UP) ? (dir + 3) % MAX_ORIENT : dir;
				facetOrient = rowLoc[orientLine][side][g].rotation;
				rowLoc[orientLine][side][g].rotation =
					(facetOrient >= 3) ?
					(rotOrientRowP3[facetOrient - 3][direction] + 3) % MAX_ORIENT :
					rotOrientRowP3[facetOrient][direction];
			}
	}

	void reverseCRD2(int orientLine) {
		PyraminxLoc temp;

		for (int g = 0; g < size; g++) {
			temp = rowLoc[orientLine][g & 1][g / 2];
			rowLoc[orientLine][g & 1][g / 2] =
				rowLoc[orientLine][((g & 1) == 1) ? 0 : 1][size - 1 - g / 2];
			rowLoc[orientLine][((g & 1) == 1) ? 0 : 1][size - 1 - g / 2] =
				temp;
		}
	}

	void reverseCRD3(int len, int orientLine) {
		PyraminxLoc temp;

		for (int g = 0; g < len; g++) {
			temp = rowLoc[orientLine][g & 1][len - ((g + 1) / 2)];
			rowLoc[orientLine][g & 1][len - ((g + 1) / 2)] =
				rowLoc[orientLine][g & 1][g / 2];
			rowLoc[orientLine][g & 1][g / 2] = temp;
		}
	}

	void writeCRD2(int view, int dir, int h, int orientLine) {
		int i, j, s, faceOnView, face, position, v = view;
		PyraminxCRD crdTemp = new PyraminxCRD(0, 0, 0);

	    switch (dir) {
		case TOP:
		case BOTTOM:
		    for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++) {
			    crdTemp.column = h;
			    crdTemp.row = g;
			    crdTemp.diagonal = h + g + side;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    facetLoc[face][position] =
				    rowLoc[orientLine][side][g];
			    drawTriangle(v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN),
				    toPosition(crdTemp), 0);
			}
		    break;
		case RIGHT:
		case LEFT:
		    for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++) {
			    crdTemp.column = g;
			    crdTemp.row = h;
			    crdTemp.diagonal = h + g + side;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    facetLoc[face][position] =
				    rowLoc[orientLine][side][g];
			    drawTriangle(face, position, 0);
			}
		    break;
		default:
		    /* dir == TR || dir == BL */
		    faceOnView = (h < size) ? DOWN : UP;
		    i = (faceOnView == UP) ? size - 1 : 0;
		    j = h % size;
		    for (int g = 0; g < size; g++) {
			for (int side = 0; side < MAX_SIDES; side++) {
			    s = (((side == UP) && (faceOnView == UP)) ||
				    ((side == DOWN) && (faceOnView == DOWN)))
				    ? DOWN : UP;
			    crdTemp.column = i;
			    crdTemp.row = j;
			    crdTemp.diagonal = i + j + s;
			    face = v * MAX_SIDES +
				    ((crdTemp.diagonal >= size) ? UP : DOWN);
			    position = toPosition(crdTemp);
			    facetLoc[face][position] =
				    rowLoc[orientLine][side][g];
			    drawTriangle(face, position, 0);
			    if (side == DOWN) {
				if (faceOnView == UP) {
				    if (j == size - 1) {
					v = (v == UP) ? DOWN : UP;
					j = 0;
				    } else
					++j;
				} else { /* FACE == DOWN */
				    if (j == 0) {
					v = (v == UP) ? DOWN : UP;
					j = size - 1;
				    } else
					--j;
				}
			    }
			}
			if (faceOnView == UP)
			    --i;
			else /* faceOnView == DOWN */
			    ++i;
		    } break;
	    }
	}

	void writeCRD3(int view, int faceOnView, int dir, int h, int len,
			int orientLine) {
		int g, i, j, s, face, position, v = view, faceView = faceOnView;
		PyraminxCRD crdTemp = new PyraminxCRD(0, 0, 0);

	    switch (dir) {
		case TOP:
		case BOTTOM:
		    for (int k = 0; k <= len; k++)
			for (int side = 0; side < MAX_SIDES; side++)
			    if ((side == DOWN) || k < len) {
				g = (faceView == UP) ? size - 1 - k : k;
				s = (((side == UP) && (faceView == DOWN)) ||
					((side == DOWN) && (faceView == UP)))
					? UP : DOWN;
				crdTemp.column = h;
				crdTemp.row = g;
				crdTemp.diagonal = h + g + s;
				face = v * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				facetLoc[face][position] =
					rowLoc[orientLine][side][k];
				drawTriangle(face, position, 0);
			    }
		    break;
		case RIGHT:
		case LEFT:
		    for (int k = 0; k <= len; k++)
			for (int side = 0; side < MAX_SIDES; side++)
			    if ((side == DOWN) || k < len) {
				g = (faceView == UP) ? size - 1 - k : k;
				s = (((side == UP) && (faceView == DOWN)) ||
					((side == DOWN) && (faceView == UP)))
					? UP : DOWN;
				crdTemp.column = g;
				crdTemp.row = h;
				crdTemp.diagonal = h + g + s;
				face = v * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				facetLoc[face][position] =
					rowLoc[orientLine][side][k];
				drawTriangle(face, position, 0);
			    }
		    break;
		default:
		    /* dir == TR || dir == BL */
		    faceView = (h < size) ? DOWN : UP;
		    i = (faceView == UP) ? size - 1 : 0;
		    j = h % size;
		    for (int k = 0; k <= len; k++) {
			for (int side = 0; side < MAX_SIDES; side++)
			    if ((side == DOWN) || k < len) {
				s = (((side == UP) && (faceView == UP)) ||
					((side == DOWN) && (faceView == DOWN)))
					? DOWN : UP;
				crdTemp.column = i;
				crdTemp.row = j;
				crdTemp.diagonal = i + j + s;
				face = v * MAX_SIDES +
					((crdTemp.diagonal >= size) ? UP : DOWN);
				position = toPosition(crdTemp);
				facetLoc[face][position] =
					rowLoc[orientLine][side][k];
				drawTriangle(face, position, 0);
				if (side == DOWN) {
				    if (faceView == UP) {
					if (j == size - 1) {
					    v = (v == UP) ? DOWN : UP;
					    j = 0;
					} else
					    ++j;
				    } else { /* FACE == DOWN */
					if (j == 0) {
					    v = (v == UP) ? DOWN : UP;
					    j = size - 1;
					} else
					    --j;
				    }
				}
			    }
			if (faceView == UP)
			    --i;
			else	/* faceView == DOWN */
			    ++i;
		    } break;
	    }
	}

	void movePieces(int face, PyraminxCRD crdRotate, int direction, int style) {
		boolean reverseRotate, bound;
		int newSide, newView, rotate, h, newH, faceOnView, len;
		int newFace, newDirection, l = 0;
		int view = face / MAX_SIDES;
		int side = crdRotate.diagonal - crdRotate.column - crdRotate.row;
		int dir = direction;

		if (style == PyraminxInterface.PERIOD2) {
			/* Period 2 Slide rows */
			h = positionCRD(dir, crdRotate.column, crdRotate.row, side);
			if (sticky && (h % 4 == 1 || h % 4 == 2)) {
				bound = true;
				l = 0;
				if (h % 4 == 2)
					h = h - 1;
			} else
				bound = false;
			do {
				readCRD2(view, dir, h, 0);
				for (int k = 1; k <= 2; k++) {
					rotate = slideNextRowP2[side][dir % 3].rotation;
					if (dir == TR || dir == BL) {
						newView = view;
						newSide = (side == UP) ? DOWN : UP;
						newH = 2 * size - 1 - h;
						reverseRotate = false;
					} else if (rotate == 0) {
						newView = (view == UP) ? DOWN : UP;
						newSide = side;
						newH = h;
						reverseRotate = false;
					} else { /* rotate == 3 */
						newView = (view == UP) ? DOWN : UP;
						newSide = (side == UP) ? DOWN : UP;
						newH = size - 1 - h;
						reverseRotate = true;
					}
					if (k != 2)
						readCRD2(newView, dir, newH, k);
					rotateCRD2(rotate, k - 1);
					if (reverseRotate)
						reverseCRD2(k - 1);
					writeCRD2(newView, dir, newH, k - 1);
					view = newView;
					h = newH;
					side = newSide;
				}
				l++;
				h++;
			} while (bound && l < 2);
		} else { /* style == PyraminxInterface.PERIOD3 */
			faceOnView = (crdRotate.diagonal >= size) ? UP : DOWN;
			h = positionCRD(dir, crdRotate.column, crdRotate.row, side);
			bound = false;
			if (dir == TR || dir == BL) {
				if (h == size)
					rotateFace(view, DOWN, (dir == TR) ? CCW : CW);
				else if (h == size - 1)
					rotateFace(view, UP, (dir == TR) ? CW : CCW);
				else if (sticky)
					bound = true;
			} else if (crdRotate.column == 0 && faceOnView == DOWN &&
				  (dir == TOP || dir == BOTTOM))
				rotateFace((view == UP) ? DOWN : UP, DOWN,
				 (dir == TOP || dir == LEFT) ? CCW : CW);
			else if (crdRotate.column == size - 1 && faceOnView == UP &&
				 (dir == TOP || dir == BOTTOM))
				rotateFace((view == UP) ? DOWN : UP, UP,
					  (dir == BOTTOM || dir == RIGHT) ? CCW : CW);
			else if (crdRotate.row == 0 && faceOnView == DOWN &&
				 (dir == RIGHT || dir == LEFT))
				rotateFace((view == UP) ? DOWN : UP, UP,
					  (dir == BOTTOM || dir == RIGHT) ? CCW : CW);
			else if (crdRotate.row == size - 1 && faceOnView == UP &&
				 (dir == RIGHT || dir == LEFT))
				rotateFace((view == UP) ? DOWN : UP, DOWN,
				 (dir == TOP || dir == LEFT) ? CCW : CW);
			else if (sticky)
				bound = true;
			/* Slide rows */
			if (bound) {
				l = 0;
				if (dir == TR || dir == BL)
					h = (faceOnView == UP) ? size + 1 : 0;
				else
					h = (faceOnView == UP) ? 0 : 1;
			}
			do {
				len = length(faceOnView, dir, h);
				readCRD3(view, faceOnView, dir, h, len, 0);
				for (int k = 1; k <= 3; k++) {
					newView = (slideNextRowP3[faceOnView][dir].viewChanged)
						? ((view == UP) ? DOWN : UP) : view;
					newFace = slideNextRowP3[faceOnView][dir].face;
					newDirection = slideNextRowP3[faceOnView][dir].direction;
					reverseRotate = slideNextRowP3[faceOnView][dir].reverse;
					newH = size - 1 - h;
					if (faceOnView == UP) {
						switch (dir) {
						case TR:
						case RIGHT:
						    newH = 2 * size - 1 - h;
						    break;
						case BOTTOM:
						    newH = h;
						    break;
						case BL:
						    newH = h - size;
						    break;
						default:
						    break;
						}
					} else {
						if (dir == TOP)
							newH = size + h;
						else if (dir == TR)
							newH = h;
					}
					if (k != 3)
						readCRD3(newView, newFace, newDirection, newH, len, k);
					rotateCRD3(faceOnView, dir, len, k - 1);
					if (reverseRotate)
						reverseCRD3(len, k - 1);
					writeCRD3(newView, newFace, newDirection, newH, len, k - 1);
					view = newView;
					faceOnView = newFace;
					dir = newDirection;
					h = newH;
				}
				h++;
				l++;
			} while (bound && l < size - 1);
		}
	}

	void moveControlCb(int face, int direction, int style) {
		PyraminxCRD crdTemp = new PyraminxCRD(-1, -1, -1);
		int faceOnView = face % MAX_SIDES;

		if (sticky) {
			if (style == PyraminxInterface.PERIOD2)
				for (int i = 0; i < 3; i++) {
					if (direction == TR || direction == BL)
					{
						crdTemp.column = 0;
						crdTemp.row = 3 * i / 2;
						crdTemp.diagonal = crdTemp.row + i % 2;
						movePieces(face, crdTemp,
							direction, style);
					} else {
						crdTemp.column = 3 * i / 2;
						crdTemp.row = 3 * i / 2;
						crdTemp.diagonal = crdTemp.row + crdTemp.column;
						movePieces(face, crdTemp,
							direction, style);
					}
					((PyraminxFrame) frame).callback(
						PyraminxInterface.ACTION_CONTROL);
				}
			else /* (style == PyraminxInterface.PERIOD3) */
				for (int i = 0; i < 2; i++) {
					if (direction == TR || direction == BL) {
						crdTemp.column = 1 + faceOnView;
						crdTemp.row = 1 + faceOnView;
						crdTemp.diagonal = crdTemp.row +
							crdTemp.column + i;
						movePieces(face, crdTemp,
							direction, style);
					} else {
						crdTemp.column = i + 2 * faceOnView;
						crdTemp.row = i + 2 * faceOnView;
						crdTemp.diagonal = crdTemp.row +
							crdTemp.column + faceOnView;
						movePieces(face, crdTemp,
							direction, style);
					}
					((PyraminxFrame) frame).callback(
						PyraminxInterface.ACTION_CONTROL);
				}
		} else {
			for (int i = 0; i < size; i++) {
				if (direction == TR || direction == BL) {
					if (style == PyraminxInterface.PERIOD2) {
						crdTemp.column = 0;
						crdTemp.row = i;
						crdTemp.diagonal = crdTemp.row;
						movePieces(face, crdTemp,
							direction, style);
					} else {
						crdTemp.column = faceOnView * (size - 1);
						crdTemp.row = i;
						crdTemp.diagonal = crdTemp.column + crdTemp.row + faceOnView;
						movePieces(face, crdTemp,
							direction, style);
					}
				} else {
					crdTemp.column = i;
					crdTemp.row = size - 1 - i;
					crdTemp.diagonal = crdTemp.column + crdTemp.row + faceOnView;
					movePieces(face, crdTemp, direction, style);
				}
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_CONTROL);
			}
		}
	}

	void movePuzzle(int face, int position, int direction, int style,
			boolean control) {
		if (control) {
			moveControlCb(face, direction, style);
		} else {
			movePieces(face, toCRD(face, position), direction, style);
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_MOVED);
		}
		undo.setMove(direction, style, (control) ? 1 : 0,
			face, position);
		redo.flushMoves();
		if (((PyraminxFrame) frame).getToggleSound())
			((PyraminxFrame) frame).playMoveAudio();
	}

	void movePuzzleDelay(int face, int position, int direction, int style,
			boolean control, int fast) {
		movePuzzle(face, position, direction, style, control);
		try {
			Thread.sleep(64 * delay / fast);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	void selectPieces() {
		if (!nextControl && !practice && checkSolved())
			moveNoPieces();

		else if (nextFace >= 0) {
			int count;

			if (period == PyraminxInterface.BOTH) {
				currentStyle = (nextShift) ?
					PyraminxInterface.PERIOD3 :
					PyraminxInterface.PERIOD2;
			} else {
				if (nextControl && nextShift)
					currentStyle = (period ==
						PyraminxInterface.PERIOD3) ?
						PyraminxInterface.PERIOD2 :
						PyraminxInterface.PERIOD3;
				else
					currentStyle = (period ==
						PyraminxInterface.PERIOD2) ?
						PyraminxInterface.PERIOD2 :
						PyraminxInterface.PERIOD3;
			}
			// next sets currentDirection
			count = checkMoveDir(currentFace,
				toCRD(currentFace, currentPosition),
				nextFace, toCRD(nextFace, nextPosition),
				currentStyle);
		    switch (count) {
			case 1:
			    narrowSelection(currentStyle);
			    movePuzzle(currentFace, currentPosition,
				    currentDirection, currentStyle,
				    nextControl);
			    started = true;
			    if (!nextControl && checkSolved()) {
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_SOLVED);
			    }
			    break;
			case 2:
			    /* count == 3 too annoying */
			    ((PyraminxFrame) frame).callback(
				    PyraminxInterface.ACTION_AMBIGUOUS);
			    break;
			case 0:
			    moveNoPieces();
			    break;
			default:
			    break;
		    }
		}
	}

	int checkMoveDir(int face1, PyraminxCRD crd1,
			int face2, PyraminxCRD crd2, int style) {
		// sets currentDirection
		int which = -1, count = 0;

		if (face1 == face2) {
			if (crd1.column == crd2.column) {
				which = 0;
				count++;
			}
			if (crd1.row == crd2.row) {
				which = 1;
				count++;
			}
			if (crd1.diagonal == crd2.diagonal) {
				which = 2;
				count++;
			}
			if (count == 1)
				switch (which) {
				case 0:	/* COLUMN */
					currentDirection =
						(crd2.row > crd1.row) ?
						BOTTOM : TOP;
					break;
				case 1:	/* ROW */
					currentDirection =
						(crd2.column > crd1.column) ?
						RIGHT : LEFT;
					break;
				case 2:	/* DIAGONAL */
					currentDirection =
						(crd2.column > crd1.column) ?
						TR : BL;
					break;
				default:
				}
			if (!vertical && face1 >= MAX_SIDES &&
					currentDirection > LEFT)
				currentDirection =
					(currentDirection + MAX_SIDES) %
					MAX_FACES;
		} else {
			if (crd1.diagonal >= size)
				crd1.diagonal = 2 * size - 1 -
					crd1.diagonal;
			if (crd2.diagonal >= size)
				crd2.diagonal = 2 * size - 1 -
					crd2.diagonal;
			if (style == 2) {
				if (crd1.column == crd2.column) {
					which = 0;
					count++;
				}
				if (face1 / 2 == face2 / 2) {
					if (crd1.row == crd2.row) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == crd2.diagonal) {
						which = 2;
						count++;
					}
				} else {
					if (crd1.row == size - 1 -
							crd2.row) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == size - 1 -
							crd2.diagonal) {
						which = 2;
						count++;
					}
				}
			} else /* if (style == 3) */ {
				if (face1 == 0 && face2 == 1) {
					if (crd1.column == size - 1 -
							crd2.row) {
						which = 0;
						count++;
					}
					if (crd1.row == size - 1 -
							crd2.column) {
						which = 1;
						count++;
					}
				} else if (face1 == 0 && face2 == 2) {
					if (crd1.row == size - 1 -
							crd2.diagonal) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == size - 1 -
							crd2.row) {
						which = 2;
						count++;
					}
				} else if (face1 == 0 && face2 == 3) {
					if (crd1.column == size - 1 -
							crd2.diagonal) {
						which = 0;
						count++;
					}
					if (crd1.diagonal == crd2.column) {
						which = 2;
						count++;
					}
				} else if (face1 == 1 && face2 == 0) {
					if (crd1.column == size - 1 -
							crd2.row) {
						which = 0;
						count++;
					}
					if (crd1.row == size - 1 -
							crd2.column) {
						which = 1;
						count++;
					}
				} else if (face1 == 1 && face2 == 2) {
					if (crd1.column == crd2.diagonal) {
						which = 0;
						count++;
					}
					if (crd1.diagonal == size - 1 -
							crd2.column) {
						which = 2;
						count++;
					}
				} else if (face1 == 1 && face2 == 3) {
					if (crd1.row == crd2.diagonal) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == crd2.row) {
						which = 2;
						count++;
					}
				} else if (face1 == 2 && face2 == 0) {
					if (crd1.row == size - 1 -
							crd2.diagonal) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == size - 1 -
							crd2.row) {
						which = 2;
						count++;
					}
				} else if (face1 == 2 && face2 == 1) {
					if (crd1.column == size - 1 -
							crd2.diagonal) {
						which = 0;
						count++;
					}
					if (crd1.diagonal == crd2.column) {
						which = 2;
						count++;
					}
				} else if (face1 == 2 && face2 == 3) {
					if (crd1.column == size - 1 -
							crd2.row) {
						which = 0;
						count++;
					}
					if (crd1.row == size - 1 -
							crd2.column) {
						which = 1;
						count++;
					}
				} else if (face1 == 3 && face2 == 0) {
					if (crd1.column == crd2.diagonal) {
						which = 0;
						count++;
					}
					if (crd1.diagonal == size - 1 -
							crd2.column) {
						which = 2;
						count++;
					}
				} else if (face1 == 3 && face2 == 1) {
					if (crd1.row == crd2.diagonal) {
						which = 1;
						count++;
					}
					if (crd1.diagonal == crd2.row) {
						which = 2;
						count++;
					}
				} else if (face1 == 3 && face2 == 2) {
					if (crd1.column == size - 1 -
							crd2.row) {
						which = 0;
						count++;
					}
					if (crd1.row == size - 1 -
							crd2.column) {
						which = 1;
						count++;
					}
				}
			}
			if (count == 1) {
				switch (which) {
				case 0:	/* COLUMN */
					if (directionNextFace[BOTTOM][face1] == face2) {
						currentDirection = BOTTOM;
					} else if (directionNextFace[TOP][face1] == face2) {
						currentDirection = TOP;
					} else {
						currentDirection = (generator.nextInt(2) == 0) ? BOTTOM :
							TOP;
					}
					break;
				case 1: /* ROW */
					if (directionNextFace[LEFT][face1] == face2) {
						currentDirection = LEFT;
					} else if (directionNextFace[RIGHT][face1] == face2) {
						currentDirection = RIGHT;
					} else {
						currentDirection = (generator.nextInt(2) == 0) ? LEFT :
							RIGHT;
					}
					break;
				case 2: /* DIAGONAL */
					if (directionNextFace[BL][face1] == face2) {
						currentDirection = BL;
					} else if (directionNextFace[TR][face1] == face2) {
						currentDirection = TR;
					} else {
						currentDirection = (generator.nextInt(2) == 0) ? BL : TR;
					}
					break;
				default:
					count = 0;
				}
			}
			if (count >= 3)
			/* Very ambiguous but lets not ignore due to face change */
				count = 2;
		}
		return count;
	}

	void narrowSelection(int style) {
		// sets currentFace, crd, currentDirection
		if (style == PyraminxInterface.PERIOD2) {
			if (currentDirection == CW)
				currentDirection = TR;
			else if (currentDirection == CCW)
				currentDirection = BL;
		} else { /* style == PyraminxInterface.PERIOD3 */
			if (currentDirection == CW || currentDirection == CCW) {
				crd.diagonal = size - ((crd.diagonal < size) ?
					UP : DOWN);
				currentDirection = ((currentDirection == CW &&
					crd.diagonal == size) ||
					(currentDirection == CCW &&
					crd.diagonal != size)) ? TR : BL;
				currentFace = ((currentFace + 1) % 2) + 2 *
					(currentFace / 2);
				crd.row = size - 1;
				crd.column = 0;
			}
		}
	}

	int positionToPiece(int positionX, int positionY) { // sets crd
		int offsetXY, modI, modJ, side, view;
		int x = positionX - puzzleOffset.x;
		int y = positionY - puzzleOffset.y;

		if (vertical && y > viewLength - 1) {
			y -= (viewLength - 1);
			view = DOWN;
		} else if (!vertical && x > viewLength - 1) {
			x -= (viewLength - 1);
			view = DOWN;
		} else
			view = UP;
		if (x <= 0 || y <= 0 ||
				x >= faceLength + delta ||
				y >= faceLength + delta)
			return -1;
		else if (x + y > faceLength + delta - 1)
			offsetXY = 2 * delta - 1;
		else
			offsetXY = delta;
		crd.column = (x - offsetXY) / (facetLength + delta);
		crd.row = (y - offsetXY) / (facetLength + delta);
		modI = (x - offsetXY) % (facetLength + delta);
		modJ = (y - offsetXY) % (facetLength + delta);
		side = (modI + modJ > facetLength + 1) ? UP : DOWN;
		if (!vertical && view == DOWN) {
			crd.row = size - crd.row - 1;
			crd.column = size - crd.column - 1;
			side = (side == UP) ? DOWN : UP;
		}
		crd.diagonal = crd.row + crd.column + side;
		return ((view == UP) ? DOWN : UP) * MAX_SIDES +
			((crd.diagonal >= size) ? UP : DOWN);
	}

	boolean positionPieces(int x, int y, int style, int direction) {
		// sets currentFace, crd, currentDirection
		if ((currentFace = positionToPiece(x, y)) == -1)
			return false;
		currentDirection = direction;
		if (!vertical && currentFace >= MAX_SIDES &&
				currentDirection < MAX_ORIENT)
			currentDirection = (currentDirection + MAX_ORIENT / 2) %
				MAX_ORIENT;
		narrowSelection(style); // sets currentFace, crd, currentDirection
		return true;
	}

	void movePuzzleInput(int x, int y, int direction, boolean shift,
			boolean control) {
		int style;

		if (period == PyraminxInterface.BOTH) {
			style = (shift) ? PyraminxInterface.PERIOD3 :
				PyraminxInterface.PERIOD2;
		} else {
			if (control && shift)
				style = (period == PyraminxInterface.PERIOD3) ?
					PyraminxInterface.PERIOD2 :
					PyraminxInterface.PERIOD3;
			else
				style = (period == PyraminxInterface.PERIOD2) ?
					PyraminxInterface.PERIOD2 :
					PyraminxInterface.PERIOD3;
		}
		if (!practice && !control && checkSolved()) {
			moveNoPieces();
			return;
		}
		if (!positionPieces(x, y, style, direction)) // sets currentFace, crd, currentDirection
			return;
		movePuzzle(currentFace, toPosition(crd), currentDirection,
			 style, control);
		if (!control && checkSolved()) {
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_SOLVED);
		}
	}

	void resetPieces() {
		solve.setFlag(false);
		sizeSize = size * size;
		facetLoc = new PyraminxLoc[MAX_FACES][sizeSize];
		startLoc = new PyraminxLoc[MAX_FACES][sizeSize];
		rowLoc = new PyraminxLoc[3][MAX_SIDES][size];
		faceLoc = new PyraminxLoc[MAX_FACES][sizeSize];
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				startLoc[face][position] = new PyraminxLoc(face, TOP);
				facetLoc[face][position] = new PyraminxLoc(face, TOP);
			}
		for (int orientLine = 0; orientLine < 3; orientLine++)
			for (int side = 0; side < MAX_SIDES; side++)
				for (int i = 0; i < size; i++)
					rowLoc[orientLine][side][i] =
						new PyraminxLoc(0, 0);
		for (int side = 0; side < MAX_SIDES; side++)
			for (int i = 0; i < sizeSize; i++)
				faceLoc[side][i] = new PyraminxLoc(0, 0);
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentFace = -1;
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void resizePieces() {
		facetLength = faceLength / size - delta - 1;
		for (int i = 0; i <= 3; i++)
			for (int j = 0; j < MAX_SIDES; j++) {
				triangleListX[j][i] = triangleUnit[j][i].x *
					facetLength + ((i > 0) ?
					triangleListX[j][i - 1] : 0);
				triangleListY[j][i] = triangleUnit[j][i].y *
					facetLength + ((i > 0) ?
					triangleListY[j][i - 1] : 0);
			}
		offsetList[DOWN] = new Point(0, 0);
		offsetList[UP] = new Point(facetLength + 2, facetLength + 2);
		letterList[DOWN] = new Point(facetLength / 4 - 3,
			facetLength / 4 + 5);
		letterList[UP] = new Point(3 * facetLength / 4,
			3 * facetLength / 4 + 5);
		sideOffset = 3 * size / 4;
		orientLineLength = facetLength / 4;
		orientDiagLength = java.lang.Math.max(orientLineLength - 3, 0);
	}

	void practicePieces() {
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_PRACTICE);
	}

	void randomizePieces() {
		if (practice)
			practicePieces();
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_RESET);
		randomPaints = java.lang.Math.min(sizeSize * 3,
			MAX_ITERATIONS) + generator.nextInt(2);
		if (sticky)
			randomPaints /= 3;
		repaint();
	}

	void RandomizingPieces() {
		int position, randomDirection, style;

		int face = generator.nextInt(MAX_FACES);
		if (period == PyraminxInterface.BOTH)
			style = generator.nextInt(PyraminxInterface.MAX_MODES - 1) +
				PyraminxInterface.PERIOD2;
		else
			style = period;
		if (sticky) {
			if (style == PyraminxInterface.PERIOD2) {
				if (generator.nextInt(3) == 2) {
					position = (generator.nextInt(2) == 1) ? 9 : 6;
					randomDirection = (generator.nextInt(2) == 1) ?
						TR : BL;
				} else {
					position = (generator.nextInt(2) == 1) ? 6 : 0;
					if (generator.nextInt(2) == 1)
						randomDirection = (generator.nextInt(2) == 1)
							? LEFT : RIGHT;
					else
						randomDirection = (generator.nextInt(2) == 1)
							? TOP : BOTTOM;
				}
			} else { /* style == PyraminxInterface.PERIOD3 */
				position = 6;
				randomDirection = generator.nextInt(6);
			}
		} else { /* (!sticky) */
			randomDirection = generator.nextInt(MAX_ORIENT);
			position = generator.nextInt(sizeSize);
			if (period == PyraminxInterface.BOTH)
				style = generator.nextInt(PyraminxInterface.BOTH);
			else
				style = period;
		}
		movePuzzle(face, position, randomDirection, style, false);
		randomPaints--;
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			// ignore
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_RANDOMIZE);
			started = false;
			currentFace = -1;
			if (checkSolved())
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_SOLVED);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
	}

	public boolean getCheat() {
		return cheat;
	}

	boolean setPuzzleSize(int newValue) {
		if (size != newValue &&
				newValue >= PyraminxInterface.MIN_FACETS) {
			size = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePeriod(int newValue) {
		if (period != newValue &&
				newValue >= PyraminxInterface.PERIOD2) {
			period = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleOrient(boolean newValue) {
		if (orient != newValue) {
			orient = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleSticky(boolean newValue) {
		if (sticky != newValue) {
			sticky = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePractice(boolean newValue) {
		if (practice != newValue) {
			practice = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((PyraminxFrame) frame).callback(PyraminxInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((PyraminxFrame) frame).callback(PyraminxInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((PyraminxFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		int tempLength;

		coreSize = new Point(getSize().width, getSize().height);
		delta = 4;
		vertical = (getSize().height >= getSize().width);
		if (vertical)
			tempLength = java.lang.Math.min(getSize().height / 2,
				getSize().width);
		else
			tempLength = java.lang.Math.min(getSize().height,
				getSize().width / 2);
		facetLength = java.lang.Math.max((tempLength - delta + 1) /
			size, 0);
		faceLength = size * facetLength;
		viewLength = faceLength + delta + 3;
		if (vertical) {
			puzzleSize = new Point(viewLength - 1,
				2 * viewLength - delta - 2);
		} else {
			puzzleSize = new Point(2 * viewLength - delta - 2,
				viewLength - 1);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		checkPieces();
		undo = new PyraminxStack();
		redo = new PyraminxStack();
		solve = new PyraminxSolve(frame, this);
		resetPieces();
		generator = new Random(System.nanoTime());
		slideNextRowP2[0][0] = new PyraminxLoc(2, 0);
		slideNextRowP2[0][1] = new PyraminxLoc(1, 3);
		slideNextRowP2[0][2] = new PyraminxLoc(2, 3);
		slideNextRowP2[1][0] = new PyraminxLoc(2, 0);
		slideNextRowP2[1][1] = new PyraminxLoc(0, 3);
		slideNextRowP2[1][2] = new PyraminxLoc(2, 3);
		slideNextRowP3[0][0] =
			new PyraminxNextRowP3(true, UP, TR, false);
		slideNextRowP3[0][1] =
			new PyraminxNextRowP3(true, UP, TOP, false);
		slideNextRowP3[0][2] =
			new PyraminxNextRowP3(false, UP, BOTTOM, true);
		slideNextRowP3[0][3] =
			new PyraminxNextRowP3(false, UP, RIGHT, true);
		slideNextRowP3[0][4] =
			new PyraminxNextRowP3(true, DOWN, RIGHT, true);
		slideNextRowP3[0][5] =
			new PyraminxNextRowP3(true, DOWN, TR, true);
		slideNextRowP3[1][0] =
			new PyraminxNextRowP3(false, DOWN, LEFT, true);
		slideNextRowP3[1][1] =
			new PyraminxNextRowP3(true, UP, LEFT, true);
		slideNextRowP3[1][2] =
			new PyraminxNextRowP3(true, UP, BL, true);
		slideNextRowP3[1][3] =
			new PyraminxNextRowP3(true, DOWN, BL, false);
		slideNextRowP3[1][4] =
			new PyraminxNextRowP3(true, DOWN, BOTTOM, false);
		slideNextRowP3[1][5] =
			new PyraminxNextRowP3(false, DOWN, TOP, true);
		triangleUnit[DOWN][0] = new Point(0, 0);
		triangleUnit[DOWN][1] = new Point(1, 0);
		triangleUnit[DOWN][2] = new Point(-1, 1);
		triangleUnit[DOWN][3] = new Point(0, -1);
		triangleUnit[UP][0] = new Point(1, 1);
		triangleUnit[UP][1] = new Point(-1, 0);
		triangleUnit[UP][2] = new Point(1, -1);
		triangleUnit[UP][3] = new Point(0, 1);
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		drawAllPieces();
		solve.init();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if ((currentFace = positionToPiece(x, y)) != -1) { //sets crd
			currentPosition = toPosition(crd);
			if (currentPosition >= sizeSize) {
				currentFace = -1;
				return;
			}
			if (control || practice || !checkSolved()) {
				selectPaint = true;
				repaint();
				return;
			}
			currentFace = -1;
		}
		currentFace = -1;
	}

	void releasePuzzle(boolean shift, boolean control, int x, int y) {
		if (currentFace == -1)
			return;
		nextShift = shift;
		nextControl = control;
		nextFace = positionToPiece(x, y); // sets crd
		nextPosition = toPosition(crd);
		releasePaint = true;
		repaint();
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizePieces();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle() {
		this.stringSave = ((PyraminxFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, sizeValue, periodValue, moves;
			boolean orientValue, stickyValue, practiceValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			sizeValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (sizeValue >= PyraminxInterface.MIN_FACETS) {
				for (i = this.size; i < sizeValue; i++) {
					((PyraminxFrame) frame).callback(
						PyraminxInterface.ACTION_INCREMENT);
				}
				for (i = this.size; i > sizeValue; i--) {
					((PyraminxFrame) frame).callback(
						PyraminxInterface.ACTION_DECREMENT);
				}
			} else {
				showMessage(source +
					" corrupted: size " + sizeValue +
					" should be at least " +
					PyraminxInterface.MIN_FACETS);
				return;
			}
			lineData = token[++line];
			periodValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (periodValue >= PyraminxInterface.PERIOD2 &&
					periodValue <= PyraminxInterface.BOTH) {
				if (periodValue != this.period) {
					int reason = periodValue -
						PyraminxInterface.PERIOD2 +
						PyraminxInterface.ACTION_PERIOD2;

					((PyraminxFrame) frame).callback(reason);
				}
			} else {
				showMessage(source +
					" corrupted: period " + periodValue +
					" should be even and between " +
					PyraminxInterface.PERIOD2 +
					" and " + PyraminxInterface.BOTH +
					" inclusive");
				return;
			}
			lineData = token[++line];
			stickyValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.sticky != stickyValue)
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_STICKY);
			lineData = token[++line];
			orientValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orient != orientValue)
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_ORIENTIZE);
			lineData = token[++line];
			practiceValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practice != practiceValue)
				((PyraminxFrame) frame).callback(
					PyraminxInterface.ACTION_PRACTICE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_RESTORE);
			setStartPosition();
			allPiecesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": size " + sizeValue +
				", period " + periodValue + ", sticky " + stickyValue +
				", orient " + orientValue +
				", practice " + practiceValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found while reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("size").append(symbol).append(space)
			.append(size).append(newline);
		sb.append("period").append(symbol).append(space)
			.append(period).append(newline);
		sb.append("sticky").append(symbol).append(space)
			.append((sticky) ? 1 : 0).append(newline);
		sb.append("orient").append(symbol).append(space)
			.append((orient) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
			.append((practice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((PyraminxFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Can not write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentStyle = undo.readMoveStyle();
		currentControl = (undo.readMoveControl() != 0);
		currentFace = undo.readMoveFace();
		currentPosition = undo.readMovePosition();
		redo.setMove(currentDirection, currentStyle, (currentControl) ?
			1 : 0, currentFace, currentPosition);
		// move in opposite direction
		currentDirection = (currentDirection < MAX_ORIENT) ?
			(currentDirection + MAX_ORIENT / 2) % MAX_ORIENT :
			3 * MAX_ORIENT - currentDirection;
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_UNDO);
		} /* else currentDirection = currentFace = -1; */
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentStyle = redo.readMoveStyle();
		currentControl = (redo.readMoveControl() != 0);
		currentFace = redo.readMoveFace();
		currentPosition = redo.readMovePosition();
		undo.setMove(currentDirection, currentStyle, (currentControl) ?
			1 : 0, currentFace, currentPosition);
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((PyraminxFrame) frame).callback(
				PyraminxInterface.ACTION_REDO);
		} /* else currentDirection = currentFace = -1; */
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetPieces();
		drawAllPieces();
		allPiecesPaint = true;
		repaint();
	}

	void solvePuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		if (((size <= 3 || sticky) && period == PyraminxInterface.PERIOD2) ||
				(size <= 4 && period == PyraminxInterface.PERIOD3)) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solvePieces();
		} else {
			showMessage("Auto-solver: sorry, only implemented for"
				+ NL + "size < 4 (or sticky mode set) where period = 2 or"
				+ NL + "size < 5 where period = 3."
				+ NL);
		}
	}

	void incrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_INCREMENT);
	}

	void decrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		if (size <= PyraminxInterface.MIN_FACETS)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_DECREMENT);
	}

	void period2ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_PERIOD2);
	}

	void period3ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_PERIOD3);
	}

	void bothModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_BOTH);
	}

	void orientizePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_ORIENTIZE);
	}

	void stickyPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_STICKY);
	}

	void practicePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		practicePieces();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void printPositions() {
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < sizeSize; position++)
			{
				int pos = startLoc[face][position].face;
				int rot = startLoc[face][position].rotation;

				if (pos < 100)
					System.out.print(" ");
				if (pos < 10 && pos >= 0)
					System.out.print(" ");
				System.out.print(pos + "," + rot);
				if (position % size == size - 1) {
					System.out.println();
				} else {
					System.out.print(" ");
				}
			}
			System.out.println();
		}
	}

	void printFacets() {
		for (int face = 0; face < MAX_SIDES; face++) {
			int square_1 = 0, sqrt = 1;

			for (int position = 0; position < sizeSize;
					position++) {
				System.out.print(
					facetLoc[face][position].face + " " +
					facetLoc[face][position].rotation +
					"  ");
				if (position == square_1) {
					System.out.println();
					++sqrt;
					square_1 = sqrt * sqrt - 1;
				}
			}
			System.out.println();
		}
		System.out.println();
	}

	void printFace() {
		for (int g = 0; g < size; g++) {
			for (int h = 0; h < size - g; h++)
				for (int side = 0; side < MAX_SIDES; side++)
				  if (side == 0 || h < size - g - 1)
				    System.out.print(
					faceLoc[side][h + g * size].face + " " +
					faceLoc[side][h + g * size].rotation +
					"  ");
			System.out.println();
		}
		System.out.println();
	}


	void printRow2(int orientLine) {
		System.out.println("Row " + orientLine + ":");
		for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++)
				System.out.print(
					rowLoc[orientLine][side][g].face + " " +
					rowLoc[orientLine][side][g].rotation +
					"  ");
		System.out.println();
	}

	void printRow3(int len, int orientLine) {
		System.out.println("Row " + orientLine + ":");
		for (int g = 0; g < size; g++)
			for (int side = 0; side < MAX_SIDES; side++)
				if (side == 0 || g < len)
				    System.out.print(
					rowLoc[orientLine][side][g].face + " " +
					rowLoc[orientLine][side][g].rotation +
					"  ");
		System.out.println();
	}

	void getStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				startLoc[face][position].face =
					facetLoc[face][position].face;
				startLoc[face][position].rotation =
					facetLoc[face][position].rotation;
			}
	}

	void setStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				facetLoc[face][position].face =
					startLoc[face][position].face;
				facetLoc[face][position].rotation =
					startLoc[face][position].rotation;
			}
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int group = (orient) ? 2 : 1;

			++line; //read startingPosition:
			for (int face = 0; face < MAX_FACES; face++) {
				int square_1 = 0, sqrt = 1, pos = 0;
				++line;
				while (token.length > line && token[line].equals(""))
					++line;
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int position = 0; position < sizeSize; position++) {
					int num = Integer.parseInt(st[group * pos + offset].trim());

					startLoc[face][position].face = num;
					if (orient) {
						num = Integer.parseInt(st[group * pos + offset + 1].trim());
						startLoc[face][position].rotation = num;
					}
					if (position == square_1) {
						if (position < sizeSize - 1) {
							line++;
							if (line >= token.length) {
								System.out.println("Short data in StartPosition");
								return -1;
							}
							lineData = token[line];
							st =  lineData.split(space);
						}
						pos = 0;
						++sqrt;
						square_1 = sqrt * sqrt - 1;
					} else
						++pos;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData =  token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentStyle = Integer.parseInt(st[1].trim());
				currentControl = (Integer.parseInt(st[2].trim()) != 0);
				currentFace = Integer.parseInt(st[3].trim());
				currentPosition = Integer.parseInt(st[4].trim());
				{
					scanPaint = true;
					undo.setMove(currentDirection,
						currentStyle,
						(currentControl) ? 1 : 0,
						currentFace, currentPosition);
					redo.flushMoves();
					paint(g);
					if (!currentControl)
						((PyraminxFrame) frame).callback(
						  PyraminxInterface.ACTION_MOVED);
				} /* else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
				} */

			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int face = 0; face < MAX_FACES; face++) {
			int square_1 = 0, sqrt = 1;
			for (int position = 0; position < sizeSize; position++)
			{
				int temp = startLoc[face][position].face;
				if (temp < 100)
					sb.append(space);
				if (temp < 10 && temp >= 0)
					sb.append(space);
				sb.append(space).append(temp);
				if (orient) {
					temp = startLoc[face][position].rotation;
					if (temp < 10 && temp >= 0)
						sb.append(space);
					sb.append(space).append(temp);
				}
				if (position == square_1) {
					sb.append(newline);
					++sqrt;
					square_1 = sqrt * sqrt - 1;
				}
			}
			sb.append(newline);
		}
	}

	void printMoves(StringBuilder sb) {
		int direction, style, control;
		int face, position;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space).append("dir").append(space).append("style").append(space).append("control").append(space).append("face").append(space).append("position").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			style = undo.readMoveStyle();
			control = undo.readMoveControl();
			face = undo.readMoveFace();
			position = undo.readMovePosition();
			redo.setMove(direction, style, control, face,
				position);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			style = redo.readMoveStyle();
			control = redo.readMoveControl();
			face = redo.readMoveFace();
			position = redo.readMovePosition();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(style).append(space)
				.append(control).append(space)
				.append(face).append(space)
				.append(position).append(newline);
			undo.setMove(direction, style, control, face,
				position);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (currentFace != -1) {
				drawTriangle(currentFace, currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentFace != -1) {
				drawTriangle(currentFace, currentPosition, 0);
				selectPieces();
				currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (currentControl)
				moveControlCb(currentFace, currentDirection,
					currentStyle);
			else {
				movePieces(currentFace, toCRD(currentFace, currentPosition),
					currentDirection, currentStyle);
				if (stackPaint && checkSolved()) {
					((PyraminxFrame) frame).callback(
						PyraminxInterface.ACTION_SOLVED);
				}
			}
			currentDirection = currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			RandomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}

/**
 * The <code>PyraminxLoc</code> class maintains the locations of
 * the pieces.
 */

class PyraminxLoc {
	int	 face;
	int	 rotation;

	public PyraminxLoc(int face, int rotation) {
		this.face = face;
		this.rotation = rotation;
	}
}

/**
 * The <code>PyraminxCRD</code> class keeps track of columns, rows and
 * diagonals in the <code>PyraminxFrame</code> class.
 */

class PyraminxCRD {
	int	 column;
	int	 row;
	int	 diagonal;

	public PyraminxCRD(int column, int row, int diagonal) {
		this.column = column;
		this.row = row;
		this.diagonal = diagonal;
	}
}

/**
 * The <code>PyraminxNextRowP3</code> class maintains positional
 * information for rows for period 3 moves.
 */

class PyraminxNextRowP3 {
	boolean     viewChanged;
	int	 face;
	int	 direction;
	boolean     reverse;

	public PyraminxNextRowP3(boolean viewChanged, int face, int direction,
			boolean reverse) {
		this.viewChanged = viewChanged;
		this.face = face;
		this.direction = direction;
		this.reverse = reverse;
	}
}
