package com.sillycycle.bagleyd.pyraminx;

/*
 * @(#)PyraminxSolve.java
 *
 * Taken from the algorithm in The Simple Solutions to Cubik Puzzles
 *  by James G. Nourse.
 *
 * Copyright 1999 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>PyraminxSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class PyraminxSolve {
	private static JFrame frame;
	private static PyraminxCanvas pyraminx;
	//private static int counter; // debugging
	private static int returnedFace, returnedPosition;
	private static boolean returnedDir;
	private static final boolean DEBUG = false;
	private static Random generator;

	PyraminxSolve(JFrame frame, PyraminxCanvas pyraminx) {
		PyraminxSolve.frame = frame;
		PyraminxSolve.pyraminx = pyraminx;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	/* Moving Nourse's large corners translates as:
		move the whole puzzle
		undo the face move */

	boolean moveAPiece(int face, int position, int direction, int style,
			boolean control) {
		if (!getFlag()) {
			return false;
		}
		pyraminx.movePuzzleDelay(face, position, direction,
			style, control, PyraminxCanvas.DOUBLE);
		return true;
	}

	void rotateLittleCornerCW(int face) {
		moveAPiece(face, 0, ((face & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			false);
	}

	void rotateLittleCornerCCW(int face) {
		moveAPiece(face, 0, ((face & 1) == 1) ?
			PyraminxCanvas.BL : PyraminxCanvas.TR, PyraminxInterface.PERIOD3,
			false);
	}

	void rotateCornerCW(int face) {
		moveAPiece(face, 1, ((face & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			true);
		moveAPiece(face, indexToEdge(2), ((face & 1) == 1) ?
			PyraminxCanvas.BL : PyraminxCanvas.TR, PyraminxInterface.PERIOD3,
			false);
	}

	void rotateCornerCCW(int face) {
		moveAPiece(face, 1, ((face & 1) == 1) ?
			PyraminxCanvas.BL : PyraminxCanvas.TR, PyraminxInterface.PERIOD3,
			true);
		moveAPiece(face, indexToEdge(2), ((face & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			false);
	}

	static int topFaceTrans(int topFace, int face) {
		return ((topFace == 0) ? face : ((topFace == 1) ?
			PyraminxCanvas.MAX_FACES - 1 - (face + PyraminxCanvas.MAX_FACES / 2)
				% PyraminxCanvas.MAX_FACES:
			((topFace == 2) ? (face + PyraminxCanvas.MAX_FACES / 2) %
				PyraminxCanvas.MAX_FACES:
			PyraminxCanvas.MAX_FACES - 1 - face)));
	}

	void P_PLUS(int topFace) {
		/* Posterior Corner CW */
		rotateCornerCW(topFaceTrans(topFace, 0));
	}

	void B_PLUS(int topFace) {
		/* Bottom Corner CW */
		rotateCornerCW(topFaceTrans(topFace, 1));
	}

	void L_PLUS(int topFace) {
		/* Left Corner CW */
		rotateCornerCW(topFaceTrans(topFace, 2));
	}

	void R_PLUS(int topFace) {
		/* Right Corner CW */
		rotateCornerCW(topFaceTrans(topFace, 3));
	}

	void P_MINUS(int topFace) {
		/* Posterior Corner CCW */
		rotateCornerCCW(topFaceTrans(topFace, 0));
	}

	void B_MINUS(int topFace) {
		/* Bottom Corner CCW */
		rotateCornerCCW(topFaceTrans(topFace, 1));
	}

	void L_MINUS(int topFace) {
		/* Left Corner CCW */
		rotateCornerCCW(topFaceTrans(topFace, 2));
	}

	void R_MINUS(int topFace) {
		/* Right Corner CCW */
		rotateCornerCCW(topFaceTrans(topFace, 3));
	}

	void rotateWholeCW(int face) {
		moveAPiece(face, 1, ((face & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			true);
	}

	void rotateWholeCCW(int face) {
		moveAPiece(face, 1, ((face & 1) == 1) ?
			PyraminxCanvas.BL : PyraminxCanvas.TR, PyraminxInterface.PERIOD3,
			true);
	}

	void B_PLUS_WHOLE(int topFace) {
		 /* Bottom Corner CW */
		 rotateWholeCW(topFaceTrans(topFace, 1));
	}

	void B_MINUS_WHOLE(int topFace) {
		 /* Bottom Corner CCW */
		rotateWholeCCW(topFaceTrans(topFace, 1));
	}

	void rotateEdgeRightPos(int face, int position) {
		moveAPiece(face, position, PyraminxCanvas.RIGHT,
			PyraminxInterface.PERIOD2, false);
	}

	void rotateEdgeTopPos(int face, int position) {
		moveAPiece(face, position, PyraminxCanvas.TOP,
			PyraminxInterface.PERIOD2, false);
	}

	void rotateEdgeTRPos(int face, int position) {
		moveAPiece(face, position, PyraminxCanvas.TR,
			PyraminxInterface.PERIOD2, false);
	}

	void rotateEdgeRight(int face) {
		rotateEdgeRightPos(face, indexToCenter(0));
	}

	void rotateEdgeTop(int face) {
		rotateEdgeTopPos(face, indexToCenter(0));
	}

	void rotateEdgeTR(int face) {
		rotateEdgeTRPos(face, indexToCenter(2));
	}

	int indexToCorner(int cornerIndex) {
		switch (cornerIndex) { /* Pyraminx -> 0, 4, 8 */
		case 0:
			return 0;
		case 1:
			return (pyraminx.size - 1) * (pyraminx.size - 1);
		case 2:
			return pyraminx.size * pyraminx.size - 1;
		default:
			System.out.println("corner unknown " + cornerIndex);
		}
		return -1;
	}

	int cornerToIndex(int corner) {
		int temp; /* Pyraminx -> {0, -1, -1, -1, 1, -1, -1, -1, 2} */

		if (corner == 0)
			return 0;
		temp = (pyraminx.size - 1) * (pyraminx.size - 1) - corner;
		if (temp == 0)
			return 1;
		else if (temp < 0) {
			temp = pyraminx.size * pyraminx.size - 1 - corner;
			if (temp == 0)
				return 2;
		}
		return -1;
	}

	int indexToEdge(int edgeIndex) {
		switch (edgeIndex) { /* Pyraminx -> 1, 3, 6 */
		case 0:
			return (pyraminx.size >> 1) * (pyraminx.size >> 1)
				+ (((pyraminx.size & 1) == 0) ? 1 : 0);
		case 1:
			return (pyraminx.size >> 1) * (pyraminx.size >> 1)
				+ 2 * (pyraminx.size >> 1)
				- (((pyraminx.size & 1) == 0) ? 1 : 0);
		case 2:
			return pyraminx.size * pyraminx.size - pyraminx.size;
		default:
				System.out.println("edge unknown " + edgeIndex);
		}
		return -1;
	}

	int edgeToIndex(int edge) {
		int temp; /* Pyraminx -> {-1, 0, -1, 1, -1, -1, 2, -1, -1} */

		temp = (pyraminx.size >> 1) * (pyraminx.size >> 1)
			+ 2 * (pyraminx.size >> 1)
			- (((pyraminx.size & 1) == 0) ? 1 : 0)
			- edge;
		if (temp == 0)
			return 1;
		else if (temp > 0) {
			temp = (pyraminx.size >> 1) * (pyraminx.size >> 1)
				+ (((pyraminx.size & 1) == 0) ? 1 : 0)
				- edge;
			if (temp == 0)
				return 0;
		} else if (temp < 0) {
			temp = pyraminx.size * pyraminx.size - pyraminx.size
				- edge;
			if (temp == 0)
				return 2;
		}
		return -1;
	}

	int indexToCenter(int centerIndex) {
		switch (centerIndex) { /* Pyraminx -> 2, 5, 7 */
		case 0:
			return 2;
		case 1:
			return (pyraminx.size - 1) * (pyraminx.size - 1) + 1;
		case 2:
			return pyraminx.size * pyraminx.size - 2;
		default:
			System.out.println("center unknown " + centerIndex);
		}
		return -1;
	}

	int centerToIndex(int center) {
		int temp; /* Pyraminx -> {-1, -1, 0, -1, -1, 1, -1, 2, -1} */

		if (center == 2)
			return 0;
		temp = (pyraminx.size - 1) * (pyraminx.size - 1) + 1 - center;
		if (temp == 0)
			return 1;
		else if (temp < 0) {
			temp = pyraminx.size * pyraminx.size - 2 - center;
			if (temp == 0)
				return 2;
		}
		return -1;
	}

	static final int[/*PyraminxCanvas.MAX_FACES*/][/*3*/] edgeMateFace = {
		{2, 3, 1},
		{3, 2, 0},
		{0, 1, 3},
		{1, 0, 2}
	};

	/* This is impossible to do with Drag and Drop as implemented.
	   Use the keypad.  Turn the little corners so that the
	   colors match the adjacent pieces all around. */
	void orientLittleCorners() {
		int face, currentColor;

		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			currentColor = pyraminx.facetLoc[face][indexToCorner(0)].face;
			if (currentColor != pyraminx.facetLoc[face][indexToCenter(0)].face) {
				if (currentColor == pyraminx.facetLoc[PyraminxCanvas.MAX_FACES -
						face - 1][indexToCenter(2)].face) {
					rotateLittleCornerCCW(face);
				} else {
					rotateLittleCornerCW(face);
				}
			}
		}
	}

	int findFaceColor(int face) {
		int[] colors = new int[3];
		int color, cornerFace;

		switch (face) {
		case 0:
			colors[0] = pyraminx.facetLoc[1][indexToCenter(0)].face;
			colors[1] = pyraminx.facetLoc[3][indexToCenter(1)].face;
			colors[2] = pyraminx.facetLoc[2][indexToCenter(2)].face;
			break;
		case 1:
			colors[0] = pyraminx.facetLoc[0][indexToCenter(0)].face;
			colors[1] = pyraminx.facetLoc[2][indexToCenter(1)].face;
			colors[2] = pyraminx.facetLoc[3][indexToCenter(2)].face;
			break;
		case 2:
			colors[0] = pyraminx.facetLoc[3][indexToCenter(0)].face;
			colors[1] = pyraminx.facetLoc[1][indexToCenter(1)].face;
			colors[2] = pyraminx.facetLoc[0][indexToCenter(2)].face;
			break;
		case 3:
			colors[0] = pyraminx.facetLoc[2][indexToCenter(0)].face;
			colors[1] = pyraminx.facetLoc[0][indexToCenter(1)].face;
			colors[2] = pyraminx.facetLoc[1][indexToCenter(2)].face;
			break;
		default:
			System.out.println("Wrong face " + face);
		}
		for (color = 0; color < PyraminxCanvas.MAX_FACES; color++) {
			for (cornerFace = 0; cornerFace < 3; cornerFace++) {
				if (color == colors[cornerFace])
					break;
			}
			if (cornerFace == 3)
				return color;
		}
		if (DEBUG)
			System.out.println("No opposite color found!");
		return 0;
	}

	boolean checkPiece(int color, int face, int position) {
		int newFace, newPosition, positionIndex;

		positionIndex = edgeToIndex(position);
		if (positionIndex == -1) {
			positionIndex = 0;
			System.out.println("incorrect position " + position);
		}
		newFace = edgeMateFace[face][positionIndex];
		newPosition = indexToEdge(positionIndex);
		return (pyraminx.facetLoc[newFace][newPosition].face == color);
	}

	void findPiece(int color0, int color1) {
		// returns returnedFace, returnedPosition
		int faceSkip, positionIndex;

		/* Check starting face first */
		for (positionIndex = 0; positionIndex < 3; positionIndex++) {
			returnedPosition = indexToEdge(positionIndex);
			if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color0 &&
					checkPiece(color1, returnedFace, returnedPosition)) {
				return;
			}
			if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color1 &&
					checkPiece(color0, returnedFace, returnedPosition)) {
				return;
			}
		}
		faceSkip = returnedFace;
		for (returnedFace = 0; returnedFace < PyraminxCanvas.MAX_FACES; returnedFace++) {
			if (returnedFace != faceSkip) {
				for (positionIndex = 0; positionIndex < 3; positionIndex++) {
					returnedPosition = indexToEdge(positionIndex);
					if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color0 &&
							checkPiece(color1, returnedFace, returnedPosition)) {
						return;
					}
					if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color1 &&
							checkPiece(color0, returnedFace, returnedPosition)) {
						return;
					}
				}
			}
		}
		if (DEBUG)
			System.out.println("Piece " + color0 + " " + color1
				+ " not found!");
	}

	void faceCorners(int face, int faceColor) {
		int otherFace;

		if (faceColor != pyraminx.facetLoc[face][indexToCenter(0)].face) {
			if (faceColor == pyraminx.facetLoc[PyraminxCanvas.MAX_FACES - 1 - face][indexToCenter(2)].face) {
				rotateCornerCW(face);
			} else {
				rotateCornerCCW(face);
			}
		}
		otherFace = (face + 2) % 4;
		if (faceColor != pyraminx.facetLoc[face][indexToCenter(1)].face) {
			if (faceColor == pyraminx.facetLoc[otherFace][indexToCenter(0)].face) {
				rotateCornerCW(otherFace);
			} else {
				rotateCornerCCW(otherFace);
			}
		}
		otherFace = PyraminxCanvas.MAX_FACES - 1 - face;
		if (faceColor != pyraminx.facetLoc[face][indexToCenter(2)].face) {
			if (faceColor == pyraminx.facetLoc[otherFace][indexToCenter(0)].face) {
				rotateCornerCCW(otherFace);
			} else {
				rotateCornerCW(otherFace);
			}
		}
	}

	/* This is a general approach that seems unwieldy right now. */
	void faceEdge(int topFace, int position, int faceColor) {
		int otherColor, newFace = 0, newPosition = 0;

		switch (edgeToIndex(position)) {
		case 0:
			newFace = (topFace + 2) % 4;
			newPosition = position + 1;
			break;
		case 1:
			newFace = PyraminxCanvas.MAX_FACES - 1 - topFace;
			newPosition = position - 1;
			break;
		case 2:
			newFace = PyraminxCanvas.MAX_FACES - 1 - ((topFace + 2) % 4);
			newPosition = position - 1;
			break;
		default:
			System.out.println("Wrong position " + position);
		}
		otherColor = pyraminx.facetLoc[newFace][newPosition].face;
		returnedFace = topFace; /* This narrows down cases */
		findPiece(faceColor, otherColor); //returns returnedFace, returnedPosition
		System.out.println("topFace " + topFace + " position " +
			position + " returnedFace " + returnedFace +
			" returnedPosition " + returnedFace + " faceColor " +
			faceColor + " otherColor " + otherColor);
	}

	void faceEdge6(int topFace, int faceColor) {
		int otherColor, newFace, newPosition, position;

		position = indexToEdge(2);
		newFace = PyraminxCanvas.MAX_FACES - 1 - ((topFace + 2) % 4);
		newPosition = indexToCenter(1);
		otherColor = pyraminx.facetLoc[newFace][newPosition].face;
		/* newFace and newPosition variables are reused here */
		returnedFace = topFace; /* This makes narrows down cases below */
		findPiece(faceColor, otherColor); // returns returnedFace, returnedPosition
		if (returnedFace == topFace) {
			switch (edgeToIndex(returnedPosition)) {
			case 0: /* Move LT to FT */
				L_MINUS(topFace);
				B_PLUS(topFace);
				L_MINUS(topFace);
				B_MINUS(topFace);
				L_MINUS(topFace);
				break;
			case 1: /* Move FR to FT */
				R_PLUS(topFace);
				B_MINUS(topFace);
				R_PLUS(topFace);
				B_PLUS(topFace);
				R_PLUS(topFace);
				break;
			case 2: /* Move FT to FT */
				/* No move :) */
				break;
			default:
				System.out.println("Wrong returned position " +
					returnedPosition);
			}
		} else { /* Little more figuring... */
			int newIndex = edgeToIndex(returnedPosition);
			if ((returnedFace == topFaceTrans(topFace, 1) ||
					returnedFace == topFaceTrans(topFace, 2)) &&
					newIndex == 1) {
				/* Move FL to FT */
				R_MINUS(topFace);
				B_PLUS(topFace);
				R_PLUS(topFace);
			} else if ((returnedFace == topFaceTrans(topFace, 1) ||
					returnedFace == topFaceTrans(topFace, 3)) &&
					newIndex == 0) {
				/* Move FR to FT */
				L_PLUS(topFace);
				B_MINUS(topFace);
				L_MINUS(topFace);
			} else if ((returnedFace == topFaceTrans(topFace, 2) ||
					returnedFace == topFaceTrans(topFace, 3)) &&
					newIndex == 2) {
				/* Move LR to FT */
				R_MINUS(topFace);
				B_MINUS(topFace);
				R_PLUS(topFace);
			} else
				System.out.println("Wrong returned face " +
					returnedFace + " or returned position "
					+ returnedPosition);
		}
		/* Not correctly positioned... need to flip it */
		if (faceColor != pyraminx.facetLoc[topFace][position].face) {
			R_MINUS(topFace);
			B_PLUS(topFace);
			R_PLUS(topFace);
			L_PLUS(topFace);
			B_PLUS(topFace);
			L_MINUS(topFace);
		}
	}

	int pickTopFace() {
		/* Pick a face (0) and determine which color it is by looking at its
		   opposite corner.  Whatever color is not there is what this
		   face should be.  Turn big corners to match */
		int topFace, faceColor, aBottomFace;

		/* Could have hard coded at 0 but this way is more challenging */
		topFace = generator.nextInt(4);
		faceColor = findFaceColor(topFace);
		faceCorners(topFace, faceColor);

		/* faceEdge(topFace, indexToEdge(0), faceColor);
		faceEdge(topFace, indexToEdge(1), faceColor);
		faceEdge(topFace, indexToEdge(2), faceColor); */

		faceEdge6(topFace, faceColor);
		aBottomFace = PyraminxCanvas.MAX_FACES - 1 - ((topFace + 2) % 4);
		moveAPiece(aBottomFace, indexToEdge(2), ((topFace & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			true);
		faceEdge6(topFace, faceColor);
		moveAPiece(aBottomFace, indexToEdge(2), ((topFace & 1) == 1) ?
			PyraminxCanvas.TR : PyraminxCanvas.BL, PyraminxInterface.PERIOD3,
			true);
		faceEdge6(topFace, faceColor);

		return topFace;
	}

	void alignBigCorner(int topFace) {
		int frontColor, leftColor, rightColor;

		frontColor = pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToCenter(2)].face;
		leftColor = pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToCenter(1)].face;
		rightColor = pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToCenter(0)].face;

		if (pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(1)].face == frontColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(1)].face == leftColor) {
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(2)].face == leftColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(2)].face == rightColor) {
			B_PLUS_WHOLE(topFace);
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(0)].face == rightColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(0)].face == frontColor) {
			B_MINUS_WHOLE(topFace);
			return;
		}

		/* CW */
		if (pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(1)].face == rightColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(1)].face == frontColor) {
			B_PLUS(topFace);
			B_MINUS_WHOLE(topFace);
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(2)].face == frontColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(2)].face == leftColor) {
			B_PLUS(topFace);
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(0)].face == leftColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(0)].face == rightColor) {
			B_PLUS(topFace);
			B_PLUS_WHOLE(topFace);
			return;
		}

		/* CCW */
		if (pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(1)].face == leftColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(1)].face == rightColor) {
			B_MINUS(topFace);
			B_PLUS_WHOLE(topFace);
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(2)].face == rightColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(2)].face == frontColor) {
			B_MINUS(topFace);
			B_MINUS_WHOLE(topFace);
			return;
		}
		if (pyraminx.facetLoc[topFaceTrans(topFace, 3)][indexToEdge(0)].face == frontColor &&
				pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(0)].face == leftColor) {
			B_MINUS(topFace);
			return;
		}
		if (DEBUG)
			System.out.println("Could not align big corner!");
	}

	boolean checkX3Solved() {
		int face, position;
		PyraminxRot test = new PyraminxRot(0, 0);

		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++)
			for (position = 0; position < pyraminx.sizeSize; position++) {
				if (cornerToIndex(position) == -1 &&
						centerToIndex(position) == -1 &&
						edgeToIndex(position) == -1)
					continue;
				if (position == 0) {
					test.face = pyraminx.facetLoc[face][position].face;
					test.rotation = pyraminx.facetLoc[face][position].rotation;
				} else if (test.face != /* MAX_SIDES * view + face */
						pyraminx.facetLoc[face][position].face ||
						(pyraminx.orient && test.rotation !=
						pyraminx.facetLoc[face][position].rotation))
					return false;
			}
		return true;
	}

	int exactCenter() {
		int half = pyraminx.size >> 1;
		return half * half + half;
	}

	/* I could not find a reasonable solution with 1
	   and only 1 center on its proper color */
	boolean checkMiddleSolvable() {
		int face, position, count = 0;
		PyraminxRot test = new PyraminxRot(0, 0);

		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			position = 0;
			test.face = pyraminx.facetLoc[face][position].face;
			position = exactCenter();
			count += ((test.face == pyraminx.facetLoc[face][position].face) ? 1 : 0);
		}
		if (DEBUG)
			System.out.println("count " + count);
		return ((count & 1) == 0);
	}

	void bottomBigCorner(int topFace) {
		int j = 0;

		do {
			int i = 0;

			if (j > 0) {
				if (DEBUG)
					System.out.println("try again");
				moveAPiece(
					((topFace >> 1) << 1) + (((topFace & 1) == 1) ? 0 : 1), exactCenter(),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
			}
			alignBigCorner(topFace);
			while (!checkX3Solved()) {
				if (i++ > 2 || j > 2) {
					if (DEBUG)
						System.out.println("Infinite Loop detected " + i + " " + j + "!");
					return;
				}
				if (pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToCenter(2)].face ==
						pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToEdge(0)].face) {
					if (pyraminx.facetLoc[topFaceTrans(topFace, 1)][indexToCenter(0)].face ==
							pyraminx.facetLoc[topFaceTrans(topFace, 2)][indexToEdge(1)].face) {
						R_MINUS(topFace);
						B_MINUS(topFace);
						R_PLUS(topFace);
						B_MINUS(topFace);
						R_MINUS(topFace);
						B_MINUS(topFace);
						R_PLUS(topFace);
					} else {
					/* Short cut: Alternate sequence */
					/* Could be solved by using above sequence but adds extra loops. */
						R_MINUS(topFace);
						B_PLUS(topFace);
						R_PLUS(topFace);
						B_PLUS(topFace);
						R_MINUS(topFace);
						B_PLUS(topFace);
						R_PLUS(topFace);
					}
				} else {
					L_PLUS(topFace);
					B_MINUS(topFace);
					L_MINUS(topFace);
					R_MINUS(topFace);
					B_MINUS(topFace);
					R_PLUS(topFace);

					L_PLUS(topFace);
					B_MINUS(topFace);
					L_MINUS(topFace);
					R_MINUS(topFace);
					B_MINUS(topFace);
					R_PLUS(topFace);
				}
				alignBigCorner(topFace);
			}
			if (DEBUG)
				System.out.println("loops " + i);
			j++;
		} while (((pyraminx.size & 1) == 0) && !checkMiddleSolvable());
	}

	/* Size 4 pyraminx */
	/* switch 2 and 3 so can have odd and even have a parity meaning for edgesLR */
	void printFacets() {
		int face, position, square;

		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			square = 1;
			for (position = 0; position < pyraminx.sizeSize; position++) {
				System.out.print(pyraminx.facetLoc[face][position].face
					+ " "
					+ pyraminx.facetLoc[face][position].rotation
					+ "  ");
				if (position == square * square - 1) {
					System.out.println();
					++square;
				}
			}
			System.out.println();
		}
		System.out.println();
	}

	static final int edgeLRMateFace[/*PyraminxCanvas.MAX_FACES*/][/*6*/] = {
		{2, 3, 3, 2, 1, 1},
		{3, 2, 2, 3, 0, 0},
		{0, 1, 1, 0, 3, 3},
		{1, 0, 0, 1, 2, 2}
	};

	static int edgeLRMatePositionIndex[/*6*/] = {3, 2, 1, 0, 5, 4};

	int indexToEdgeLR(int edgeIndex) {
		switch (edgeIndex) { /* Size 4 Pyraminx -> 1, 3, 4, 8, 11, 13 */
		case 0:
			return 1;
		case 1:
			return 3;
		case 3:
			return (pyraminx.size - 2) * (pyraminx.size - 2);
		case 2:
			return (pyraminx.size - 1) * (pyraminx.size - 1) - 1;
		case 4:
			return (pyraminx.size - 1) * (pyraminx.size - 1) + 2;
		case 5:
			return pyraminx.size * pyraminx.size - 3;
		default:
			System.out.println("edge unknown " + edgeIndex);
		}
		return -1;
	}

	int edgeLRToIndex(int edge) {
		int temp; /* Size 4 Pyraminx -> {-1, 0, -1, 1, 3, -1, -1, -1, 2, -1, -1, 4, -1, 5, -1, -1} */

		if (edge == 1)
			return 0;
		else if (edge == 3)
			return 1;
		temp = (pyraminx.size - 1) * (pyraminx.size - 1) - 1 - edge;
		if (temp == 0)
			return 2;
		else if (temp > 0) {
			temp = (pyraminx.size - 2) * (pyraminx.size - 2) - edge;
			if (temp == 0)
				return 3;
		} else if (temp < 0) {
			if (temp == -3)
				return 4;
			temp = pyraminx.size * pyraminx.size - 3 - edge;
			if (temp == 0)
				return 5;
		}
		return -1;
	}

	boolean checkPieceLR(int color, int face, int position) {
		int newFace, newPosition, positionIndex, newPositionIndex;

		positionIndex = edgeLRToIndex(position);
		if (positionIndex == -1) {
			positionIndex = 0;
			System.out.println("position " + position + " incorrect");
		}
		newFace = edgeLRMateFace[face][positionIndex];
		newPositionIndex = edgeLRMatePositionIndex[positionIndex];
		newPosition = indexToEdgeLR(newPositionIndex);
		return (pyraminx.facetLoc[newFace][newPosition].face == color);
	}

	/* color0 is even position */
	void findPieceLR(int color0, int color1) {
		// returns returnedFace, returnedPosition
		int faceSkip, positionIndex, startPositionIndex;

		startPositionIndex = edgeLRToIndex(returnedPosition);
		/* Check starting face first */
		for (positionIndex = 0; positionIndex < 6; positionIndex++) {
			returnedPosition = indexToEdgeLR(positionIndex);
			if (((startPositionIndex + positionIndex) % 2) == 0) {
				if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color0 &&
						checkPieceLR(color1, returnedFace, returnedPosition)) {
					if (DEBUG)
						System.out.println("LRe " + returnedFace + " " + returnedPosition);
					return;
				}
			} else {
				if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color1 &&
						checkPieceLR(color0, returnedFace, returnedPosition)) {
					if (DEBUG)
						System.out.println("LRo " + returnedFace + " " + returnedPosition);
					return;
				}
			}
		}
		faceSkip = returnedFace;
		for (returnedFace = 0; returnedFace < PyraminxCanvas.MAX_FACES; returnedFace++) {
			if (returnedFace == faceSkip)
				continue;
			for (positionIndex = 0; positionIndex < 6; positionIndex++) {
				returnedPosition = indexToEdgeLR(positionIndex);
				if (((startPositionIndex + positionIndex) % 2) == 0) {
					if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color0 &&
							checkPieceLR(color1, returnedFace, returnedPosition)) {
						if (DEBUG)
							System.out.println("LR e " + returnedFace + " " + returnedPosition);
						return;
					}
				} else {
					if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color1 &&
							checkPieceLR(color0, returnedFace, returnedPosition)) {
						if (DEBUG)
							System.out.println("LR o " + returnedFace + " " + returnedPosition);
						return;
					}
				}
			}
		}
		if (DEBUG)
			System.out.println("PieceLR " + " " + color0 + " " + color1
				+ " not found!");
		printFacets();
	}

	void solve0LR0() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][0].face;
		int color1 = pyraminx.facetLoc[2][0].face;
		int startFace = 0, startPositionIndex = 0;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
				break;
			case 1:
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 4:
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve0LR1() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][0].face;
		int color1 = pyraminx.facetLoc[3][0].face;
		int startFace = 0, startPositionIndex = 1;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
				System.out.println("impossible");
				break;
			case 1:
				break;
			case 2:
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 4:
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(returnedFace, returnedPosition,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve0LR2() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[2][indexToCorner(1)].face;
		int color1 = pyraminx.facetLoc[3][indexToCorner(2)].face;
		int startFace = 2, startPositionIndex = 4;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
				System.out.println("impossible");
				break;
			case 2:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, indexToEdgeLR(4),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(4),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(4),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(4),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve1LR0() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][0].face;
		int color1 = pyraminx.facetLoc[2][0].face;
		int startFace = 0, startPositionIndex = 3;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
				System.out.println("impossible");
				break;
			case 2:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				break;
			case 4:
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, returnedPosition,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), returnedPosition,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), returnedPosition,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 4:
				System.out.println("impossible");
				break;
			case 5:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve1LR1() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][0].face;
		int color1 = pyraminx.facetLoc[3][0].face;
		int startFace = 0, startPositionIndex = 2;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
			case 3:
				System.out.println("impossible");
				break;
			case 2:
				break;
			case 4:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 4:
				System.out.println("impossible");
				break;
			case 5:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve1LR2() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[2][indexToCorner(1)].face;
		int color1 = pyraminx.facetLoc[3][indexToCorner(2)].face;
		int startFace = 2, startPositionIndex = 5;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
			case 2:
			case 3:
				System.out.println("impossible");
				break;
			case 4:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 0:
				System.out.println("impossible");
				break;
			case 1:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			case 2:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				System.out.println("impossible");
				break;
			case 5:
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve2LR1() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[2][0].face;
		int color1 = pyraminx.facetLoc[1][0].face;
		int startFace = 2, startPositionIndex = 1;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
			case 2:
			case 3:
				System.out.println("impossible");
				break;
			case 4:
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			switch (positionIndex) {
			case 1:
				break;
			case 2:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			case 4:
				break;
			}
			break;
		case 3:
			break;
		}
	}

	void solve2LR2() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[2][0].face;
		int color1 = pyraminx.facetLoc[1][0].face;
		int startFace = 2, startPositionIndex = 2;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
			case 2:
			case 3:
				System.out.println("impossible");
				break;
			case 4:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 1:
				break;
			case 3:
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			break;
		case 3:
			break;
		}
	}

	void solve0LR4() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][indexToCorner(1)].face;
		int color1 = pyraminx.facetLoc[1][indexToCorner(2)].face;
		int startFace = 0, startPositionIndex = 4;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
				System.out.println("impossible");
				break;
			case 2:
				break;
			case 4:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 5:
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 3,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(2),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(3),
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 3,
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), indexToEdgeLR(3),
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			break;
		case 3:
			break;
		}
	}

	void solve0LR5() {
		int positionIndex;
		int color0 = pyraminx.facetLoc[0][indexToCorner(1)].face;
		int color1 = pyraminx.facetLoc[1][indexToCorner(2)].face;
		int startFace = 0, startPositionIndex = 5;

		returnedFace = startFace;
		returnedPosition = indexToEdgeLR(startPositionIndex);
		findPieceLR(color0, color1);
		positionIndex = edgeLRToIndex(returnedPosition);
		if (DEBUG)
			System.out.println("findPieceLR color0=" + color0
				+ " color1=" + color1 + " : face=" + returnedFace
				+ " positionIndex= " + positionIndex);
		switch (returnedFace) {
		case 0:
			switch (positionIndex) {
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
				System.out.println("impossible");
				break;
			case 5:
				break;
			}
			break;
		case 1:
			switch (positionIndex) {
			case 0:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			case 3:
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.BL, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.TOP, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.LEFT, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, indexToEdgeLR(2),
					PyraminxCanvas.BOTTOM, PyraminxInterface.PERIOD3, false);
				moveAPiece(startFace, 1,
					PyraminxCanvas.RIGHT, PyraminxInterface.PERIOD3, false);
				moveAPiece(((startFace >> 1) << 1) + (((startFace & 1) == 1) ? 0 : 1), 1,
					PyraminxCanvas.TR, PyraminxInterface.PERIOD3, false);
				break;
			}
			break;
		case 2:
			break;
		case 3:
			break;
		}
	}

	void solveTopLevel() {
		solve0LR0();
		solve0LR1();
		solve0LR2();
	}

	void solveMiddleLevel() {
		solve1LR0();
		solve1LR1();
		solve1LR2();
	}

	void solveBottomLevel() {
		solve2LR1();
		solve2LR2();
		solve0LR4();
		solve0LR5();
	}

	int findPieceCenter(int color0, int face) {
		int faceSkip;
		int position = exactCenter();

		/* Check starting face first */
		if (pyraminx.facetLoc[face][position].face == color0) {
			if (DEBUG)
				System.out.println("Center " + face);
			return face;
		}
		faceSkip = face;
		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			if (face == faceSkip)
				continue;
			if (pyraminx.facetLoc[face][position].face == color0) {
				if (DEBUG)
					System.out.println("Center fs " + face + " " + position);
				return face;
			}
		}
		System.out.println("Piece " + color0 + " not found!");
		return -1;
	}

	void solve4CenterDown(int startFace) {
		int color0 = pyraminx.facetLoc[startFace][0].face;
		int position = exactCenter();
		int index = findPieceCenter(color0, startFace);

		if (index == 0)
			return;
		index--;
		moveAPiece(startFace, position,
			(PyraminxCanvas.BOTTOM + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.TR + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.TOP + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.LEFT + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.BOTTOM + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);

		moveAPiece(startFace, position,
			(PyraminxCanvas.RIGHT + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.BL + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.LEFT + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.TOP + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
		moveAPiece(startFace, position,
			(PyraminxCanvas.RIGHT + 2 * index) % PyraminxCanvas.COORD, PyraminxInterface.PERIOD3, false);
	}

	static int middleOrient[/*PyraminxCanvas.MAX_FACES*/] = {0, 0, 0, 0};

	void findPieceCenterOrient() {
		int rotation = pyraminx.facetLoc[0][0].rotation;
		int position = exactCenter();
		int face;
		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			middleOrient[face] =
				(pyraminx.facetLoc[face][position].rotation
				- rotation + PyraminxCanvas.COORD) % 3;
			if (DEBUG)
				System.out.println("rotation " + face + " " + middleOrient[face]);
		}
	}

	void orientShuffle(int face, int parity, int rotation) {
		int position = exactCenter();
		int i;
		for (i = 0; i < 3; i++) {
			moveAPiece(face, position,
				(PyraminxCanvas.LEFT + 3 * parity + 2 * rotation) % PyraminxCanvas.COORD,
				PyraminxInterface.PERIOD3, false);
			moveAPiece(face, position,
				(PyraminxCanvas.TOP + 3 * parity + 2 * rotation) % PyraminxCanvas.COORD,
				PyraminxInterface.PERIOD3, false);
			moveAPiece(face, position,
				(PyraminxCanvas.RIGHT + 3 * parity + 2 * rotation) % PyraminxCanvas.COORD,
				PyraminxInterface.PERIOD3, false);
			moveAPiece(face, position,
				(PyraminxCanvas.BOTTOM + 3 * parity + 2 * rotation) % PyraminxCanvas.COORD,
				PyraminxInterface.PERIOD3, false);
		}
	}

	void orientShuffleFaces(int face0, int face1) {
		int newFace0, newFace1;
		if (DEBUG)
			System.out.println("orientShuffleFaces " + face0 + " " + face1);
		if ((face0 == 0 && face1 == 1) || (face0 == 1 && face1 == 0)
				|| (face0 == 2 && face1 == 3)
				|| (face0 == 3 && face1 == 2)) {
			newFace1 = (face0 + 2) % PyraminxCanvas.MAX_FACES;
			newFace0 = (face1 + 2) % PyraminxCanvas.MAX_FACES;
			orientShuffle(newFace0, (newFace0 & 1), 0);
			orientShuffle(newFace1, (newFace1 & 1), 0);
		} else if ((face0 == 0 && face1 == 2) || (face0 == 2 && face1 == 0)
				|| (face0 == 1 && face1 == 3)
				|| (face0 == 3 && face1 == 1)) {
			if (face0 == 1 || face0 == 3) {
				newFace0 = (face0 + 3) % PyraminxCanvas.MAX_FACES;
				newFace1 = (face1 + 3) % PyraminxCanvas.MAX_FACES;
			} else {
				newFace1 = (face0 + 3) % PyraminxCanvas.MAX_FACES;
				newFace0 = (face1 + 3) % PyraminxCanvas.MAX_FACES;
			}
			orientShuffle(newFace0, (newFace0 & 1), 1);
			orientShuffle(newFace1, (newFace1 & 1), 1);
		} else if ((face0 == 0 && face1 == 3) || (face0 == 3 && face1 == 0)
				|| (face0 == 1 && face1 == 2)
				|| (face0 == 2 && face1 == 1)) {
			newFace0 = (face0 + 2) % PyraminxCanvas.MAX_FACES;
			newFace1 = (face1 + 2) % PyraminxCanvas.MAX_FACES;
			orientShuffle(newFace0, (newFace0 & 1), 2);
			orientShuffle(newFace1, (newFace1 & 1), 2);
		}
	}

	int findCenter(int start, int value) {
		int face;
		for (face = start + 1; face < PyraminxCanvas.MAX_FACES; face++) {
			if (middleOrient[face] == value)
				return face;
		}
		return -1;
	}

	void decideShuffle() {
		int face0 = findCenter(-1, 1);
		int face1 = findCenter(-1, 2);

		if (face0 >= 0 && face1 >= 0) {
			if (DEBUG)
				System.out.println("decideShuffle " + face0 + " " + face1);
			orientShuffleFaces(face0, face1);
			findPieceCenterOrient();
		}
		face0 = findCenter(-1, 1);
		face1 = findCenter(face0, 1);
		if (face0 >= 0 && face1 >= 0) {
			if (DEBUG)
				System.out.println("decideShuffle " + face0 + " " + face1);
			orientShuffleFaces(face0, face1);
			findPieceCenterOrient();
		}
		face0 = findCenter(-1, 2);
		face1 = findCenter(face0, 2);
		if (face0 >= 0 && face1 >= 0) {
			if (DEBUG)
				System.out.println("decideShuffle " + face0 + " " + face1);
			orientShuffleFaces(face0, face1);
			findPieceCenterOrient();
		}
		face0 = findCenter(-1, 1);
		face1 = findCenter(-1, 2);
		if (face0 >= 0 && face1 >= 0) {
			if (DEBUG)
				System.out.println("decideShuffle " + face0 + " " + face1);
			orientShuffleFaces(face0, face1);
		}
	}

	void solveCorners() {
		orientLittleCorners();
		if (pyraminx.size > 2) {
			bottomBigCorner(pickTopFace());
		}
		if (pyraminx.size > 3) {
			solveTopLevel();
			solveMiddleLevel();
			solveBottomLevel();
			solve4CenterDown(0);
			if (pyraminx.orient) {
				findPieceCenterOrient();
				decideShuffle();
			}
		}
	}

	static PyraminxFC[] pyraminxFC = new PyraminxFC[3];

/* 2e */
/* rotateEdge
2      0       6 e2
0      1       2 e0
1      2       6 e2
0      0       2
2      0       6
0      0       2
1      2       6
0      1       2
*/


	void findCornerEdgePositions() {
		int reference, center, counter = 0, face, corner;

		center = pyraminx.size >> 1;
		center = center * center + center;
		reference = pyraminx.facetLoc[0][center].face;
		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			for (corner = 0; corner < 3; corner++) {
				if (reference ==
						pyraminx.facetLoc[face][indexToCorner(corner)].face) {
					pyraminxFC[counter] = new PyraminxFC(face, corner);
					counter++;
					if (counter == 3)
						return;
				}
			}
		}
	}

	void printEdgeError(PyraminxFC[] pfc) {
		System.out.println("fix me "
			+ pfc[0].face + " " + pfc[0].corner + "  "
			+ pfc[1].face + " " + pfc[1].corner + "  "
			+ pfc[2].face + " " + pfc[2].corner);
	}

	void cornerEdges() {
		findCornerEdgePositions();
		switch (pyraminxFC[0].face) {
		case 0:
		    switch (pyraminxFC[0].corner) {
		    case 0:
			switch (pyraminxFC[1].face) {
			case 0:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    case 1:
				switch (pyraminxFC[2].face) {
				case 0:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					System.out.println("solved already");
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    case 2:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 1:
			    switch (pyraminxFC[1].corner) {
			    case 1:
				switch (pyraminxFC[2].face) {
				case 1:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    case 2:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    case 1:
			switch (pyraminxFC[1].face) {
			case 0:
			    switch (pyraminxFC[1].corner) {
			    case 2:
				switch (pyraminxFC[2].face) {
				case 1:
				    switch (pyraminxFC[2].corner) {
				    case 0:
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 1:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 2:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 0:
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 3:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    case 2:
			switch (pyraminxFC[1].face) {
			case 1:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    default:
			printEdgeError(pyraminxFC);
		    }
		    break;
		case 1:
		    switch (pyraminxFC[0].corner) {
		    case 0:
			switch (pyraminxFC[1].face) {
			case 1:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    case 1:
				switch (pyraminxFC[2].face) {
				case 1:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    case 2:
				switch (pyraminxFC[2].face) {
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    case 1:
			switch (pyraminxFC[1].face) {
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    case 2:
			switch (pyraminxFC[1].face) {
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 1:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 0:
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 3:
			    switch (pyraminxFC[1].corner) {
			    case 0:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 1:
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    default:
			printEdgeError(pyraminxFC);
		    }
		    break;
		case 2:
		    switch (pyraminxFC[0].corner) {
		    case 0:
			switch (pyraminxFC[1].face) {
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 1:
				switch (pyraminxFC[2].face) {
				case 2:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			case 3:
			    switch (pyraminxFC[1].corner) {
			    case 1:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    case 1:
			switch (pyraminxFC[1].face) {
			case 2:
			    switch (pyraminxFC[1].corner) {
			    case 2:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 0:
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					rotateEdgeRight(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    default:
			printEdgeError(pyraminxFC);
		    }
		    break;
		case 3:
		    switch (pyraminxFC[0].corner) {
		    case 0:
			switch (pyraminxFC[1].face) {
			case 3:
			    switch (pyraminxFC[1].corner) {
			    case 1:
				switch (pyraminxFC[2].face) {
				case 3:
				    switch (pyraminxFC[2].corner) {
				    case 2:
					rotateEdgeRight(0);
					rotateEdgeTR(0);
					rotateEdgeTop(0);
					rotateEdgeTR(0);
					break;
				    default:
					printEdgeError(pyraminxFC);
				    }
				    break;
				default:
				    printEdgeError(pyraminxFC);
				}
				break;
			    default:
				printEdgeError(pyraminxFC);
			    }
			    break;
			default:
			    printEdgeError(pyraminxFC);
			}
			break;
		    default:
			printEdgeError(pyraminxFC);
		    }
		    break;
		default:
		    printEdgeError(pyraminxFC);
		}
	}

	void
	findPieceOrder(int color0, int color1) {
		int positionIndex;
		for (returnedFace = 0; returnedFace < PyraminxCanvas.MAX_FACES; returnedFace++) {

			for (positionIndex = 0; positionIndex < 3; positionIndex++) {
				int returnedPosition = indexToEdge(positionIndex);
				if (pyraminx.facetLoc[returnedFace][returnedPosition].face == color0 &&
						checkPiece(color1, returnedFace, returnedPosition)) {
					return;
				}
			}
		}
		System.out.println("Ordered piece "
			+ color0 + " " + color1 +" not found!");
	}

	/*
	void
	solveEdgeEdgeTurn2() {
		int face = 0;

		rotateEdgeRightPos(face, 1);
		rotateEdgeTopPos(face, indexToCorner(2));
		rotateEdgeTRPos(face, 0);
		rotateEdgeTopPos(face, 0);

		rotateEdgeRightPos(face, 1);
		rotateEdgeTopPos(face, 0);
		rotateEdgeTRPos(face, 0);
		rotateEdgeTopPos(face, indexToCorner(2));
	}
	*/

	void
	solve0Edge() {
		int face = 0, otherFace = 2;
		int faceColor, otherColor, newPositionIndex;
		faceColor =  pyraminx.facetLoc[face][0].face;
		otherColor = pyraminx.facetLoc[otherFace][0].face;
		returnedFace = face;
		findPieceOrder(faceColor, otherColor);
		newPositionIndex = edgeToIndex(returnedPosition);
		switch (returnedFace) {
		case 0:
			/* done */
			break;
		case 1:
			rotateEdgeTopPos(returnedFace, indexToEdge(0));
			rotateEdgeRightPos(returnedFace, indexToEdge(0));
			rotateEdgeTopPos(returnedFace, indexToEdge(0));
			break;
		case 2:
			rotateEdgeTopPos(face, indexToEdge(0));
			rotateEdgeRightPos(face, indexToEdge(0));
			rotateEdgeTopPos(face, indexToEdge(0));
			rotateEdgeRightPos(face, indexToEdge(0));
			break;
		case 3:
			rotateEdgeRightPos(face, indexToEdge(0));
			break;
		}
		if (DEBUG)
			System.out.println("solve0Edge: "
				+ returnedFace + " " + newPositionIndex);
	}

	void
	solve1Edge() {
		int face = 0, otherFace = 3;
		int faceColor, otherColor, newPositionIndex;
		faceColor =  pyraminx.facetLoc[face][0].face;
		otherColor = pyraminx.facetLoc[otherFace][0].face;
		returnedFace = face;
		findPieceOrder(faceColor, otherColor);
		newPositionIndex = edgeToIndex(returnedPosition);
		switch (returnedFace) {
		case 0:
			/* done */
			break;
		case 1:
			rotateEdgeRightPos(face, indexToEdge(1));
			rotateEdgeTopPos(face, indexToEdge(1));
			rotateEdgeRightPos(face, indexToEdge(1));
			break;
		case 2:
			rotateEdgeTopPos(face, indexToEdge(1));
			break;
		case 3:
			rotateEdgeRightPos(face, indexToEdge(1));
			rotateEdgeTopPos(face, indexToEdge(1));
			rotateEdgeRightPos(face, indexToEdge(1));
			rotateEdgeTopPos(face, indexToEdge(1));
			break;
		}
		if (DEBUG)
			System.out.println("solve1Edge: "
				+ returnedFace + " " + newPositionIndex);
	}

	void
	solve2Edge() {
		int face = 0, otherFace = 1;
		int faceColor, otherColor, newPositionIndex;
		faceColor =  pyraminx.facetLoc[face][0].face;
		otherColor = pyraminx.facetLoc[otherFace][0].face;
		returnedFace = face;
		findPieceOrder(faceColor, otherColor);
		newPositionIndex = edgeToIndex(returnedPosition);
		if (returnedFace == 1) {
			if (DEBUG)
				System.out.println("solve2Edge");
			face = 1;
			otherFace = 3;
			faceColor =  pyraminx.facetLoc[face][0].face;
			otherColor = pyraminx.facetLoc[otherFace][0].face;
			returnedFace = face;
			findPieceOrder(faceColor, otherColor);
			newPositionIndex = edgeToIndex(returnedPosition);
			if (returnedFace == 3) {
				if (DEBUG)
					System.out.println("solve2Edge 3");
				rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeRightPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTopPos(0, pyraminx.size * pyraminx.size - 1);
				rotateEdgeRightPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				return;
			}
			otherFace = 2;
			otherColor = pyraminx.facetLoc[otherFace][0].face;
			returnedFace = face;
			findPieceOrder(faceColor, otherColor);
			newPositionIndex = edgeToIndex(returnedPosition);
			if (returnedFace == 2) {
				if (DEBUG)
					System.out.println("solve2Edge 2");
				rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTopPos(0, pyraminx.size * pyraminx.size - 1);
				rotateEdgeRightPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTopPos(0, pyraminx.size * pyraminx.size - 1);
				return;
			}
			if (DEBUG)
				System.out.println("solve2Edge opp");
			rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
			rotateEdgeTopPos(0, indexToEdge(1));
			rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
			rotateEdgeTopPos(0, indexToEdge(1));
		}
	}

	void
	solve3Edge() {
		int face = 2, otherFace = 3;
		int faceColor, otherColor, newPositionIndex;
		faceColor =  pyraminx.facetLoc[face][0].face;
		otherColor = pyraminx.facetLoc[otherFace][0].face;
		returnedFace = face;
		findPieceOrder(faceColor, otherColor);
		newPositionIndex = edgeToIndex(returnedPosition);
		if (returnedFace == 3) {
			if (DEBUG)
				System.out.println("solve3Edge");
			otherFace = 1;
			otherColor = pyraminx.facetLoc[otherFace][0].face;
			returnedFace = face;
			findPieceOrder(faceColor, otherColor);
			newPositionIndex = edgeToIndex(returnedPosition);
			if (returnedFace == 1) {
				if (DEBUG)
					System.out.println("solve3Edge 1");
				rotateEdgeTRPos(2, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTopPos(2, 0);
				rotateEdgeRightPos(2, 0);
				rotateEdgeTopPos(2, 0);
				return;
			}
			otherFace = 3;
			otherColor = pyraminx.facetLoc[otherFace][0].face;
			returnedFace = face;
			findPieceOrder(faceColor, otherColor);
			newPositionIndex = edgeToIndex(returnedPosition);
			if (returnedFace == 3) {
				if (DEBUG)
					System.out.println("solve3Edge 2");
				rotateEdgeTRPos(2, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeRightPos(2, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTopPos(2, pyraminx.size * pyraminx.size - 1);
				rotateEdgeRightPos(2, (pyraminx.size - 1) * (pyraminx.size - 1));
				return;
			}
		}
	}

	void
	solve4Edge() {
		int face = 1, otherFace = 2;
		int faceColor, otherColor, newPositionIndex;
		faceColor =  pyraminx.facetLoc[face][0].face;
		otherColor = pyraminx.facetLoc[otherFace][0].face;
		returnedFace = face;
		findPieceOrder(faceColor, otherColor);
		newPositionIndex = edgeToIndex(returnedPosition);
		if (returnedFace == 2) {
			if (DEBUG)
				System.out.println("solve4Edge");
			face = 2;
			otherFace = 3;
			faceColor =  pyraminx.facetLoc[face][0].face;
			otherColor = pyraminx.facetLoc[otherFace][0].face;
			returnedFace = face;
			findPieceOrder(faceColor, otherColor);
			newPositionIndex = edgeToIndex(returnedPosition);
			if (returnedFace == 2) {
				if (DEBUG)
					System.out.println("solve4Edge 2");
				rotateEdgeTopPos(0, pyraminx.size * pyraminx.size - 1);
				rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeRightPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				rotateEdgeTRPos(0, (pyraminx.size - 1) * (pyraminx.size - 1));
				return;
			}
		}
	}

	void
	solveEdgeEdgeTurn() {
		solve0Edge();
		solve1Edge();
		solve2Edge();
		solve3Edge();
		solve4Edge();
	}

	static int[] cornerFace = new int[4];
	static int[] centerFace = new int[4];

	void
	solveCenterEdgeRow(int face, Boolean dir) {
		int i;

		if (dir) {
			for (i = 0; i < 2; i++) {
				rotateEdgeTRPos(face, 0);
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeRightPos(face, indexToCorner(1));
			}
		} else {
			for (i = 0; i < 2; i++) {
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTRPos(face, 0);
			}
		}
	}

	void
	solveCenterEdgeColumn(int face, Boolean dir) {
		int i;

		if (dir) {
			for (i = 0; i < 2; i++) {
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeTRPos(face, 0);
			}
		} else {
			for (i = 0; i < 2; i++) {
				rotateEdgeTRPos(face, 0);
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTopPos(face, indexToCorner(2));
			}
		}
	}

	void
	solveCenterEdgeDiagonal(int face, Boolean dir) {
		int i;

		if (dir) {
			for (i = 0; i < 2; i++) {
				rotateEdgeTRPos(face, indexToCorner(2));
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeTRPos(face, indexToCorner(2));
				rotateEdgeRightPos(face, indexToCorner(1));
			}
		} else {
			for (i = 0; i < 2; i++) {
				rotateEdgeRightPos(face, indexToCorner(1));
				rotateEdgeTRPos(face, indexToCorner(2));
				rotateEdgeTopPos(face, indexToCorner(2));
				rotateEdgeTRPos(face, indexToCorner(2));
			}
		}
	}

	int
	findCenterPositions(int corner) {
		int face, count = 0;

		returnedFace = PyraminxCanvas.MAX_FACES;
		returnedDir = false;
		for (face = 0; face < PyraminxCanvas.MAX_FACES; face++) {
			cornerFace[face] =
				pyraminx.facetLoc[face][indexToCorner(corner)].face;
			centerFace[face] =
				pyraminx.facetLoc[face][indexToCenter(corner)].face;
			if (cornerFace[face] == centerFace[face]) {
				count++;
				returnedFace = face;
			}
		}
		if (count != 1) {
			returnedFace = PyraminxCanvas.MAX_FACES;
		} else {
			face = 0;
			if (returnedFace == 0 || returnedFace == 1)
				face = 2;
			returnedDir = (centerFace[face] == cornerFace[face + 1]);
			if ((returnedFace & 1) == 1)
				returnedDir = !returnedDir;
		}
		return count;
	}

	void solveMiddleRow() {
		int count;

		count = findCenterPositions(1);
		if (count == 0) {
			if (DEBUG)
				System.out.println("all row 4");
			solveCenterEdgeRow(0, true);
			count = findCenterPositions(1);
		}
		if (count == 1) {
			if (DEBUG)
				System.out.println("row "
					+ returnedFace + " " + returnedDir);
			solveCenterEdgeRow(returnedFace, returnedDir);
			return;
		}
		if (DEBUG)
			System.out.println("row done"); /* count == 4 */
	}

	void solveMiddleColumn() {
		int count;

		count = findCenterPositions(2);
		if (count == 0) {
			if (DEBUG)
				System.out.println("all column 4");
			solveCenterEdgeColumn(0, true);
			count = findCenterPositions(2);
		}
		if (count == 1) {
			if (DEBUG)
				System.out.println("column "
					+ returnedFace + " " + returnedDir);
			solveCenterEdgeColumn(returnedFace, returnedDir);
			return;
		}
		if (DEBUG)
			System.out.println("column done"); /* count == 4 */
	}

	void solveMiddleDiagonal() {
		int count;

		count = findCenterPositions(0);
		if (count == 0) {
			if (DEBUG)
				System.out.println("all diagonal 4");
			solveCenterEdgeDiagonal(0, true);
			count = findCenterPositions(0);
		}
		if (count == 1) {
			if (DEBUG)
				System.out.println("diagonal "
					+ returnedFace + " " + returnedDir);
			solveCenterEdgeDiagonal(returnedFace, returnedDir);
			return;
		}
		if (DEBUG)
			System.out.println("diagonal done"); /* count == 4 */
	}

	void solveCenterEdgeTurn() {
		solveMiddleRow();
		solveMiddleColumn();
		solveMiddleDiagonal();
	}

	void solveEdges() {
		cornerEdges();
		if (pyraminx.size > 2) {
			if (pyraminx.sticky || pyraminx.size % 2 != 0) {
				solveEdgeEdgeTurn();
			}
			solveCenterEdgeTurn();
		}
	}

	/* This procedure coordinates the solution process. */
	void solvePieces() {
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_RESET);
		if (!pyraminx.checkSolved()) {
			if (pyraminx.period == PyraminxInterface.PERIOD2)
				solveEdges();
			else if (pyraminx.period == PyraminxInterface.PERIOD3)
				solveCorners();
		}
		((PyraminxFrame) frame).callback(
			PyraminxInterface.ACTION_COMPUTED);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solvePieces();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}

class PyraminxRot {
	int	face;
	int	rotation;

	public PyraminxRot(int face, int rotation) {
		this.face = face;
		this.rotation = rotation;
	}
}

class PyraminxFC {
	int	face;
	int	corner;

	public PyraminxFC(int face, int corner) {
		this.face = face;
		this.corner = corner;
	}
}
